"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import importlib.metadata
from importlib.metadata import EntryPoint

import typer
from typer.models import TyperInfo

log = logging.getLogger(__name__)


def load_command_plugins(app: typer.Typer) -> None:
    """
    Dynamically load external plugins and add them as subcommands.
    """

    # Discover plugins via entry points defined in pyproject.toml
    for entry_point in importlib.metadata.entry_points(group="pp_terminal.commands"):
        try:
            plugin = entry_point.load()
            if not isinstance(plugin, typer.Typer):
                raise RuntimeError('not a Typer app')

            group = _get_app_group_from_entry_point(app, entry_point)
            if group.typer_instance is None:
                raise RuntimeError('missing instance')

            group.typer_instance.add_typer(plugin)
            log.debug('loaded plugin command "%s" into group "%s"', entry_point.name, group.name)
        except Exception as e: # pylint: disable=broad-exception-caught
            log.error("failed to load plugin %s, ignoring: %s", entry_point.name, e)


def _get_app_group_from_entry_point(app: typer.Typer, entry_point: EntryPoint) -> TyperInfo:
    for group in app.registered_groups:
        if group.name == entry_point.name.split('.', 1)[0] and group.typer_instance is not None:
            return group

    return TyperInfo(app, name="")



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import sqlite3
import os
import logging
from pathlib import Path
from types import SimpleNamespace

from ppxml2db.ppxml2db import PortfolioPerformanceXML2DB
from ppxml2db import dbhelper
from ppxml2db import ppxml2db_init

from pp_terminal.exceptions import InputError

log = logging.getLogger(__name__)
logging.getLogger('ppxml2db.dbhelper').setLevel(logging.INFO)  # reducing some "noise"


DB_NAME_IN_MEMORY = ':memory:'


class Ppxml2dbWrapper:
    _connection: sqlite3.Connection
    _schema_path: str
    _cursor: sqlite3.Cursor

    def __init__(self, dbname: str = DB_NAME_IN_MEMORY) -> None:
        self._schema_path = os.path.dirname(dbhelper.__file__) + '/'

        try:
            dbhelper.init('sqlite', dbname)  # type: ignore
            if dbhelper.db is None:
                raise RuntimeError('could not establish database connection')

            self._connection = dbhelper.db  # @todo db connection is global state!
            self._cursor = self._connection.cursor()
            self._install()
        except Exception as e:
            raise RuntimeError('error during database initialization for ' + dbname) from e

    def open(self, file: Path) -> None:
        try:
            conv = PortfolioPerformanceXML2DB(file.open(mode='rb'))  # type: ignore
            conv.iterparse()  # type: ignore

            self._connection.commit()

            self._validate()
        except FileNotFoundError as e:
            raise e
        except Exception as e:
            raise InputError('unable to import the Portfolio Performance xml file "' + file.name + '" (is it saved as "XML with ids"?)') from e

    def close(self) -> None:
        self._connection.close()  # database is deleted

    @property
    def connection(self) -> sqlite3.Connection:
        return self._connection

    def _install(self) -> None:
        # Check if tables already exist (for cached databases)
        try:
            self._validate()
            log.debug('Database tables already exist, skipping initialization scripts')
        except Exception:  # pylint: disable=broad-exception-caught
            # Tables don't exist, create them using official ppxml2db schema
            self._create_tables()

    def _create_tables(self) -> None:
        # Use ppxml2db_init.main() to ensure schema compatibility
        args = SimpleNamespace(dbtype='sqlite', recreate=False)
        current_dir = os.getcwd()
        try:
            os.chdir(self._schema_path)
            ppxml2db_init.main(args)  # type: ignore
            log.debug('ppxml2db schema initialized')
        finally:
            os.chdir(current_dir)

    def _validate(self) -> None:
        self._cursor.execute("select value as client_version from property where name = 'version'")
        row = self._cursor.fetchone()
        if row is None:
            raise RuntimeError('missing client version in xml file')



import sys
import argparse
import logging

from .version import __version__
from . import dbhelper


SCHEMA = [
    "account",
    "account_attr",
    "security",
    "security_attr",
    "security_event",
    "security_prop",
    "latest_price",
    "price",
    "watchlist",
    "watchlist_security",
    "xact",
    "xact_unit",
    "xact_cross_entry",
    "taxonomy",
    "taxonomy_category",
    "taxonomy_data",
    "taxonomy_assignment",
    "taxonomy_assignment_data",
    "dashboard",
    "property",
    "bookmark",
    "attribute_type",
    "config_set",
    "config_entry",
]


_log = logging.getLogger(__name__)


def drop_table(table):
    sql = "DROP TABLE IF EXISTS %s" % table
    if args.dbtype == "pgsql":
        sql += " CASCADE"
    dbhelper.execute_dml(sql)


def main(args):
    for table in SCHEMA:
        with open(table + ".sql") as f:
            sql = f.read()
        if args.recreate:
            drop_table(table)

        if args.dbtype == "pgsql":
            sql = sql.replace(
                "INTEGER NOT NULL PRIMARY KEY",
                "SERIAL NOT NULL PRIMARY KEY"
            )

        dbhelper.executescript(sql)


if __name__ == "__main__":
    argp = argparse.ArgumentParser(description="Initialize database for ppxml2db")
    argp.add_argument("db", help="output DB (filename/connect string)")
    argp.add_argument("--dbtype", choices=("sqlite", "pgsql"), default="sqlite", help="select database type")
    argp.add_argument("--recreate", action="store_true", help="delete existing tables")
    argp.add_argument("--debug", action="store_true", help="enable debug logging")
    argp.add_argument("--dry-run", action="store_true", help="don't commit changes to DB")
    argp.add_argument("--version", action="version", version="%(prog)s " + __version__)
    args = argp.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    dbhelper.init(args.dbtype, args.db)

    main(args)

    if not args.dry_run:
        dbhelper.commit()



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from pp_terminal.output.strategy import OutputFormat, OutputStrategy, RichOutputStrategy, CsvOutputStrategy, \
    JsonOutputStrategy
from pp_terminal.output.strategy_excel import ExcelOutputStrategy


def create_strategy(output_format: OutputFormat) -> OutputStrategy:
    match output_format:
        case OutputFormat.TABLE:
            return RichOutputStrategy()
        case OutputFormat.CSV:
            return CsvOutputStrategy()
        case OutputFormat.JSON:
            return JsonOutputStrategy()
        case OutputFormat.EXCEL:
            return ExcelOutputStrategy()

    raise NotImplementedError('output format "' + output_format + '" not supported yet')



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from datetime import datetime
from typing import cast

import typer
from pp_terminal.data.cost_basis import calculate_total_cost_basis
from pp_terminal.data.filters import filter_by_security
from pp_terminal.output.column_utils import normalize_columns
from pp_terminal.utils.config import Config
from pp_terminal.utils.helper import footer
from pp_terminal.output.strategy import OutputStrategy, Console
from pp_terminal.domain.portfolio import Portfolio
from pp_terminal.domain.portfolio_snapshot import PortfolioSnapshot
from pp_terminal.output.table_decorator import TableOptions
from pp_terminal.validation.engine import validate_securities, ValidationResult
from pp_terminal.utils.config import get_command_config

app = typer.Typer()
console = Console()
log = logging.getLogger(__name__)


@app.command(name="securities")
def print_securities(  # pylint: disable=too-many-locals
    ctx: typer.Context,
    by: datetime = datetime.now(),
    active: bool = False,
    in_stock: bool = False,
    columns: str | None = None
) -> None:
    """
    Show a detailed table with all securities and their IDs.
    """

    portfolio = cast(Portfolio, ctx.obj.portfolio)
    output = cast(OutputStrategy, ctx.obj.output)
    config = cast(Config, ctx.obj.config)

    if columns is None:
        config_columns = get_command_config(config, 'view.securities.columns')
        columns = ','.join(config_columns) if config_columns else 'SecurityId,Name,Wkn,Currency,Shares,Messages'

    securities = portfolio.securities
    snapshot = PortfolioSnapshot(portfolio, by)
    shares = snapshot.shares

    # Reset index to make SecurityId a column and rename columns
    df = securities.reset_index()

    if shares is not None and not shares.empty:
        shares_by_security = shares.groupby('securityId').sum()
        df = df.merge(shares_by_security, left_on='securityId', right_index=True, how='left', validate='one_to_one')
        df['shares'] = df['shares'].fillna(0.0)
    else:
        df['shares'] = 0.0

    if active and 'isRetired' in df.columns:
        df = df[~df['isRetired']]

    if in_stock:
        df = df[df['shares'] > 0.001]

    validation_results = validate_securities(portfolio, config)
    df['Messages'] = df['securityId'].map(
        lambda sid: validation_results.get(str(sid), ValidationResult.empty()).messages or ''
    )

    df['costBasis'] = df['securityId'].map(
        lambda sid: calculate_total_cost_basis(portfolio.securities_account_transactions.pipe(filter_by_security, security_id=sid))
    )

    requested_columns = [col.strip() for col in columns.split(',')]
    selected_columns = normalize_columns(requested_columns, list(df.columns), portfolio.security_attributes)

    df = df[selected_columns]
    df = df.rename(columns=portfolio.security_attributes)

    if 'isRetired' in df.columns and 'isRetired' not in columns:
        df = df.drop(columns=['isRetired'])

    df = df.sort_values(by='name') if 'name' in df.columns else df

    console.print(*output.result_table(
        df, TableOptions(
            title=f"{'Active ' if active else ''}Securities",
            caption=f"{len(df)} entries per {by.strftime("%Y-%m-%d")}",
            show_index=False,
            show_total=False
        )
    ))
    console.print(output.text(footer()), style="dim")



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""
import typer

from pp_terminal.domain.schemas import Percent


def tax_rate_callback(ctx: typer.Context, param: typer.CallbackParam, value: Percent | None) -> Percent:  # pylint: disable=unused-argument
    # 1. If provided via CLI, use it
    if value is not None:
        return value

    # 2. Try config file
    config_tax = ctx.obj.config.get('tax')
    if config_tax is not None and 'rate' in config_tax:
        return Percent(config_tax['rate'])

    # 3. Prompt user with 4. hard-coded default
    return Percent(typer.prompt(
        "Tax Rate (%)",
        type=float,
        default=0.25 * (1 + 0.055) * 100
    ))


def exemption_rate_callback(ctx: typer.Context, param: typer.CallbackParam, value: Percent | None) -> Percent:  # pylint: disable=unused-argument
    # 1. If provided via CLI, use it
    if value is not None:
        return value

    # 2. Try config file
    config_tax = ctx.obj.config.get('tax')
    if config_tax is not None and 'exemption-rate' in config_tax:
        return Percent(config_tax['exemption-rate'])

    # 3. Prompt user with 4. hard-coded default
    return Percent(typer.prompt(
        "Default Exemption Rate (%)",
        type=float,
        default=30
    ))


def tax_csv_callback(ctx: typer.Context, param: typer.CallbackParam, value: str | None) -> str | None:  # pylint: disable=unused-argument
    # 1. If provided via CLI, use it
    if value is not None:
        return value

    # 2. Try config file
    config_tax = ctx.obj.config.get('tax')
    if config_tax is not None and 'file' in config_tax:
        return str(config_tax['file'])

    return None



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from datetime import datetime
from typing import cast

import pandas as pd
import typer

from pp_terminal.domain.portfolio import Portfolio
from pp_terminal.output.column_utils import normalize_columns
from pp_terminal.data.filters import unstack_column_by_currency
from pp_terminal.exceptions import InputError
from pp_terminal.utils.helper import footer
from pp_terminal.output.strategy import OutputStrategy, Console
from pp_terminal.domain.portfolio_snapshot import PortfolioSnapshot
from pp_terminal.domain.schemas import AccountType
from pp_terminal.output.table_decorator import TableOptions
from pp_terminal.validation.engine import validate_accounts, ValidationResult
from pp_terminal.utils.config import get_command_config, Config

app = typer.Typer()
console = Console()
log = logging.getLogger(__name__)


def _prepare_df_for_display(
    df: pd.DataFrame,
    selected_columns_preunstack: list[str],
    snapshot: PortfolioSnapshot,
    unstack_balance: bool
) -> pd.DataFrame:
    """Prepare DataFrame for display with optional balance unstacking."""
    if unstack_balance:
        cols_before_unstack = set(df.columns)
        df = df.pipe(unstack_column_by_currency, column='balance', base_currency=snapshot.portfolio.base_currency)
        currency_cols = list(set(df.columns) - cols_before_unstack)
    else:
        currency_cols = []

    # Drop currency column before reset_index if it exists (to avoid collision with index level)
    if 'currency' in df.columns:
        df = df.drop(columns=['currency'])

    df = df.reset_index()

    selected_columns = []
    for col in selected_columns_preunstack:
        if col == 'balance' and currency_cols:
            selected_columns.extend(currency_cols)
        elif col in df.columns:
            selected_columns.append(col)

    df = df[selected_columns]
    df = df.rename(columns=snapshot.portfolio.account_attributes)

    if 'accountId' in df.columns:
        df = df.set_index('accountId')

    return df


def calculate_deposit_accounts_sum(snapshot: PortfolioSnapshot) -> pd.DataFrame:
    balances = (pd.merge(snapshot.portfolio.deposit_accounts, snapshot.balances, left_index=True, right_on='accountId', how="right", validate='one_to_many')
            .sort_values(by='balance'))

    balances = balances[balances['balance'] >= 0.01]
    # Drop columns that are not useful for display
    cols_to_drop = [col for col in balances.columns if col in ['referenceAccount', 'isRetired']]
    if cols_to_drop:
        balances = balances.drop(columns=cols_to_drop)
    return balances


def calculate_securities_accounts_sum(snapshot: PortfolioSnapshot) -> pd.DataFrame:
    values = (pd.merge(snapshot.portfolio.securities_accounts, snapshot.values.groupby(['accountId', 'currency']).sum(), left_index=True, right_on='accountId', how="right", validate='one_to_many')
            .sort_values(by='balance'))

    values = values[values['balance'] >= 0.01]
    # Drop columns that are not useful for display
    cols_to_drop = [col for col in values.columns if col in ['referenceAccount', 'isRetired']]
    if cols_to_drop:
        values = values.drop(columns=cols_to_drop)
    return values


@app.command(name="accounts")
def print_accounts(  # pylint: disable=too-many-locals
    ctx: typer.Context,
    type: AccountType | None = None,  # pylint: disable=redefined-builtin
    by: datetime = datetime.now(),
    columns: str | None = None
) -> None:
    """
    Show a detailed table with the current balance per deposit account.
    """

    portfolio = cast(Portfolio, ctx.obj.portfolio)
    output = cast(OutputStrategy, ctx.obj.output)
    config = cast(Config, ctx.obj.config)

    if columns is None:
        config_columns = get_command_config(config, 'view.accounts.columns')
        columns = ','.join(config_columns) if config_columns else 'AccountId,Name,Type,Balance,Messages'

    snapshot = PortfolioSnapshot(portfolio, by)

    df1 = None
    if type == AccountType.DEPOSIT or type is None:
        df1 = calculate_deposit_accounts_sum(snapshot)

    df2 = None
    if type == AccountType.SECURITIES or type is None:
        df2 = calculate_securities_accounts_sum(snapshot)

    df = pd.concat([df1, df2]) if df1 is not None or df2 is not None else None

    if df is None:
        raise InputError('invalid account type')

    # Add validation messages column
    validation_results = validate_accounts(portfolio, snapshot, config)
    account_ids = df.index.get_level_values('accountId')
    df['Messages'] = account_ids.map(
        lambda aid: validation_results.get(str(aid), ValidationResult.empty()).messages or ''
    )

    requested_columns = [col.strip() for col in columns.split(',')]

    # Available columns before unstacking - need to account for accountId which will be from the index
    available_before_unstack = list(set(df.columns) - {'balance'}) + ['accountId']
    if 'balance' in df.columns:
        available_before_unstack.append('balance')

    selected_columns_preunstack = normalize_columns(requested_columns, available_before_unstack, portfolio.account_attributes)

    df = _prepare_df_for_display(
        df, selected_columns_preunstack, snapshot,
        unstack_balance='balance' in selected_columns_preunstack and 'balance' in df.columns
    )

    console.print(*output.result_table(
        df, TableOptions(title="Balances on Accounts", caption=f"{len(df)} entries per {by.strftime("%Y-%m-%d")}", show_index=True)
    ))
    console.print(output.text(footer()), style="dim")



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime
import logging
from pathlib import Path
from typing import cast
from typing_extensions import Annotated

import typer
from pp_terminal.data.cost_basis import calculate_fifo_sell
from pp_terminal.data.filters import filter_by_account_and_security

from pp_terminal.data.tax import load_prepaid_tax_data_from_csv
from pp_terminal.exceptions import InputError
from pp_terminal.utils.helper import format_money, footer
from pp_terminal.utils.options import tax_rate_callback, tax_csv_callback
from pp_terminal.output.strategy import OutputStrategy, Console
from pp_terminal.domain.portfolio_snapshot import PortfolioSnapshot
from pp_terminal.domain.portfolio import Portfolio, get_securities_account_by_id, get_security_by_id
from pp_terminal.domain.schemas import Percent, Money, Account, Security
from pp_terminal.output.table_decorator import TableOptions

app = typer.Typer()
console = Console()
log = logging.getLogger(__name__)


def get_today() -> datetime:
    """Return today's date at midnight."""
    return datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)


def _get_available_shares(security: Security, account: Account, snapshot: PortfolioSnapshot) -> float:
    shares_available = snapshot.shares
    if shares_available is None:
        raise InputError("No share holdings found in portfolio")

    # Check for this specific account/security combination
    holding_key = None
    for key in shares_available.index:
        if key[0] == account.accountId and key[1] == security.securityId:
            holding_key = key
            break

    if holding_key is None:
        raise InputError(f"No shares of '{security.name}' found in account '{account.name}' on {snapshot.date.strftime('%Y-%m-%d')}")

    return float(shares_available[holding_key])


@app.command(name="share-sell")
def simulate_share_sell(  # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals,too-many-statements,too-many-branches
        ctx: typer.Context,
        security_id: Annotated[str, typer.Option(prompt="Security ID", prompt_required=True)],
        account_id: Annotated[str, typer.Option(help="Securities account ID", prompt="Account ID", prompt_required=True)],
        date: Annotated[datetime | None, typer.Option(formats=["%Y-%m-%d"], help="Sale date (defaults to today)", prompt="Sale date (YYYY-MM-DD)", prompt_required=False)] = None,
        tax_rate: Annotated[Percent, typer.Option(help="Your personal tax rate", min=0, max=100, callback=tax_rate_callback)] = None,  # type: ignore
        shares: Annotated[float | None, typer.Option(help="Number of shares to sell (defaults to all available shares)", min=0.0001)] = None,
        price: Annotated[Money | None, typer.Option(help="Sale price per share (defaults to latest market price)")] = None,
        tax_csv: Annotated[Path | None, typer.Option(help="CSV file with paid tax per share data", callback=tax_csv_callback)] = None
) -> None:
    """
    Simulate selling shares: calculate fees, taxes (Abgeltungssteuer + Soli), and net proceeds.
    Uses FIFO cost basis and accounts for taxes already paid.
    """
    portfolio = cast(Portfolio, ctx.obj.portfolio)
    output = cast(OutputStrategy, ctx.obj.output)

    if date is None:
        date = get_today()

    _tax_csv_data = load_prepaid_tax_data_from_csv(tax_csv, tax_rate) if tax_csv else None

    security = get_security_by_id(portfolio, security_id)
    account = get_securities_account_by_id(portfolio, account_id)
    snapshot = PortfolioSnapshot(portfolio, date)

    # Determine sale price
    if price is None:
        latest_prices = snapshot.latest_prices
        if security_id not in latest_prices.index:
            raise InputError(f"No price data available for security '{security_id}'. Please provide --price")
        sale_price = latest_prices.loc[security_id]
    elif price <= 0:
        raise InputError("Sale price must be greater than 0")
    else:
        sale_price = price

    available_shares = _get_available_shares(security, account, snapshot)
    if shares is None:
        shares = available_shares
    elif available_shares < shares - 0.0001:  # Allow small floating point errors
        raise InputError(f"Insufficient shares. Available: {available_shares:.8f}, Requested: {shares:.8f}")

    transactions = snapshot.securities_account_transactions.pipe(filter_by_account_and_security, security_id=security_id, account_id=account_id)
    fifo_lots = calculate_fifo_sell(transactions, snapshot.date, sale_price, tax_rate, shares, _tax_csv_data).reset_index()

    console.print(output.text(f"\n[bold]Security:[/bold] {security.name} ({security.wkn})"))
    console.print(output.text(f"[bold]Account:[/bold] {account.name}"))
    console.print(output.text(f"[bold]Shares:[/bold] {shares}"))
    console.print(output.text(f"[bold]Sale Date:[/bold] {date.strftime('%Y-%m-%d')}"))
    console.print(output.text(f"[bold]Sale Price (per share):[/bold] {format_money(sale_price, security.currency)}"))

    console.print(*output.result_table(
        fifo_lots[['date', 'shares', 'currency', 'purchasePrice', 'cost', 'fees', 'salePrice', 'capitalGain', 'taxableGain', 'grossProceeds', 'totalTax', 'netProceeds']],
        TableOptions(title="FIFO Lots Breakdown", show_index=False, show_total=True)
    ))

    console.print(output.warning(f'This simulation assumes all values are in security currency ({security.currency}) excl. Sparerpauschbetrag.'))
    console.print(output.text(footer()), style="dim")



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from pathlib import Path
from typing import Any, Dict

import numpy as np
import pandas as pd
from pandera.typing import DataFrame

from pp_terminal.domain.portfolio import Portfolio
from pp_terminal.domain.schemas import TransactionSchema, AccountSchema, SecuritySchema, SecurityPriceSchema, TransactionType
from pp_terminal.utils.cache import cleanup_old_cache_files, get_cache_path
from pp_terminal.utils.helper import enum_list_to_values
from .attribute_type_converter import convert_attribute_types, get_converter_column_name
from .ppxml2db_wrapper import Ppxml2dbWrapper, DB_NAME_IN_MEMORY

log = logging.getLogger(__name__)

_SCALE = 100000000
_CENTS_PER_EURO = 100
_NEGATIVE_DEPOSIT_ACCOUNT_TRANSACTION_TYPES = [
    TransactionType.TRANSFER_OUT,
    TransactionType.REMOVAL,
    TransactionType.INTEREST_CHARGE,
    TransactionType.FEES,
    TransactionType.TAXES,
    TransactionType.BUY,
]
_ATTRIBUTE_TYPE_ACCOUNT = 'name.abuchen.portfolio.model.Account'
_ATTRIBUTE_TYPE_SECURITY = 'name.abuchen.portfolio.model.Security'


class PpPortfolioBuilder:  # pylint: disable=too-few-public-methods
    _db: Ppxml2dbWrapper
    _config: Dict[str, Any]

    def __init__(self, db: Ppxml2dbWrapper | None = None, config: Dict[str, Any] | None = None):
        self._db = db if db is not None else Ppxml2dbWrapper(dbname=DB_NAME_IN_MEMORY)
        self._config = config if config is not None else {}

    def construct(self, file: Path) -> Portfolio:
        self._db.open(file)

        portfolio = Portfolio(
            accounts = self._parse_accounts(),
            transactions = self._parse_transactions(),
            securities = self._parse_securities(),
            prices = self._parse_prices(),
            attributes = {
                'accounts': self._get_attributes(_ATTRIBUTE_TYPE_ACCOUNT),
                'securities': self._get_attributes(_ATTRIBUTE_TYPE_SECURITY)
            }
        )

        portfolio.base_currency = str(self._get_property('baseCurrency'))

        self._db.close()

        return portfolio

    def _parse_securities(self) -> DataFrame[SecuritySchema]:
        security_attrs = self._get_attributes(_ATTRIBUTE_TYPE_SECURITY)

        # Build dynamic SQL with CASE statements for each attribute
        case_statements = []
        params = []
        for attr_uuid in security_attrs:
            case_statements.append(f'MAX(CASE WHEN at.id = ? THEN sa.value END) AS "{attr_uuid}"')
            params.append(attr_uuid)
            case_statements.append(f'MAX(CASE WHEN at.id = ? THEN at.converterClass END) AS "{get_converter_column_name(attr_uuid)}"')
            params.append(attr_uuid)

        sql = f"""
            select s.*,
                {', '.join(case_statements) if case_statements else '""'}
            from security as s
            left join security_attr as sa on sa."security" = s.uuid
            left join attribute_type as at on sa.attr_uuid = at.id
            group by s.uuid
        """  # nosec B608 - case_statements contain only column names with ? placeholders, data passed via params

        securities = pd.read_sql_query(sql, self._db.connection, index_col='uuid', params=params).rename_axis('securityId')
        securities = convert_attribute_types(securities, security_attrs)

        return SecuritySchema.validate(securities)

    def _parse_prices(self) -> DataFrame[SecurityPriceSchema]:
        prices = (pd.read_sql_query('select datetime(tstamp) as date, * from price', self._db.connection, index_col=['date', 'security'], parse_dates={"date": "%Y-%m-%d %H:%M:%S"}, dtype={'value': np.float64})
                          .rename(columns={'value': 'price'}))[['price']]
        prices['price'] = prices['price'] / _SCALE
        prices.index.set_names(['date', 'securityId'], inplace=True)

        return SecurityPriceSchema.validate(prices)

    def _parse_transactions(self) -> DataFrame[TransactionSchema]:
        transactions = (pd.read_sql_query("""
select datetime(x.date) as date, ifnull(xu.forex_currency, x.currency) as currency, ifnull(xu.forex_amount, x.amount)-x.fees as amount_wo_fees, x.fees, x.taxes, x.uuid, x.account, x.type, x.security, x.shares, x.acctype
from xact as x
left join xact_unit as xu on xu.xact = x.uuid and xu.type = 'GROSS_VALUE'
        """, self._db.connection, index_col=['date', 'account', 'security'], parse_dates={"date": "%Y-%m-%d %H:%M:%S"}, dtype={'amount_wo_fees': np.float64, 'shares': np.float64, 'taxes': np.float64})
                          .rename(columns={'amount_wo_fees': 'amount', 'acctype': 'accountType'}))
        transactions['shares'] = transactions['shares'] / _SCALE
        transactions['type'] = pd.Categorical(transactions['type'])
        transactions['amount'] = transactions.apply(
            lambda row: -1 if row['type'] in enum_list_to_values(_NEGATIVE_DEPOSIT_ACCOUNT_TRANSACTION_TYPES) else 1,
            axis=1
        ) * transactions['amount'] / _CENTS_PER_EURO
        transactions['taxes'] = transactions['taxes'] / _CENTS_PER_EURO
        transactions['fees'] = transactions['fees'] / _CENTS_PER_EURO
        transactions.index.set_names(['date', 'accountId', 'securityId'], inplace=True)

        return TransactionSchema.validate(transactions)

    def _parse_accounts(self) -> DataFrame[AccountSchema]:
        account_attrs = self._get_attributes(_ATTRIBUTE_TYPE_ACCOUNT)

        # Build dynamic SQL with CASE statements for each attribute
        case_statements = []
        params = []
        for attr_uuid in account_attrs:
            case_statements.append(f'MAX(CASE WHEN at.id = ? THEN aa.value END) AS "{attr_uuid}"')
            params.append(attr_uuid)
            case_statements.append(f'MAX(CASE WHEN at.id = ? THEN at.converterClass END) AS "{get_converter_column_name(attr_uuid)}"')
            params.append(attr_uuid)

        sql = f"""
            select a.*,
                {', '.join(case_statements) if case_statements else '""'}
            from account as a
            left join account_attr as aa on aa."account" = a.uuid
            left join attribute_type as at on aa.attr_uuid = at.id
            group by a.uuid
        """  # nosec B608 - case_statements contain only column names with ? placeholders, data passed via params

        accounts = (pd.read_sql_query(sql, self._db.connection, index_col='uuid', params=params)
                          .rename_axis('accountId'))

        # Convert attribute values based on converterClass
        accounts = convert_attribute_types(accounts, account_attrs)

        return AccountSchema.validate(accounts)

    def _get_property(self, name: str) -> str | None:
        cursor = self._db.connection.cursor()
        cursor.execute('select value from property where name = ?', (name, ))

        result = cursor.fetchone()
        if result is None:
            return None

        return str(result[0])

    def _get_attributes(self, entity: str) -> dict[str, str]:
        cursor = self._db.connection.cursor()
        cursor.execute("SELECT id, name FROM attribute_type WHERE target = ? AND id NOT IN ('logo')", (entity, ))

        return {str(row[0]): str(row[1]) for row in cursor.fetchall()}


class CachedPpPortfolioBuilder:  # pylint: disable=too-few-public-methods
    _config: Dict[str, Any]

    def __init__(self, config: Dict[str, Any] | None = None):
        self._config = config if config is not None else {}

    def construct(self, file: Path) -> Portfolio:
        use_cache_file = False

        try:
            cache_path = get_cache_path(file)

            if cache_path.exists():
                log.debug('Using cache database "%s"', cache_path)
                use_cache_file = True
            else:
                log.debug('Cache file not found, will create at "%s"', cache_path)

            db = Ppxml2dbWrapper(dbname=str(cache_path))
        except Exception as e:  # pylint: disable=broad-exception-caught
            log.warning(
                'Failed to initialize cache database (%s), using in-memory database',
                str(e)
            )
            db = Ppxml2dbWrapper(dbname=DB_NAME_IN_MEMORY)
            cache_path = None
            use_cache_file = False

        builder = PpPortfolioBuilder(db, self._config)

        # Override the open behavior for cache hits
        if use_cache_file:
            # Cache hit: skip XML parsing by replacing open with a no-op
            original_open = db.open
            db.open = lambda _: None  # type: ignore
            portfolio = builder.construct(file)
            db.open = original_open  # type: ignore
        else:
            # Cache miss: normal construction which will call db.open()
            portfolio = builder.construct(file)

            if cache_path is not None:
                try:
                    cleanup_old_cache_files(file)
                except Exception as e:  # pylint: disable=broad-exception-caught
                    log.warning('Failed to cleanup old cache files: %s', str(e))

        return portfolio



import sys
import argparse
import logging
from collections import defaultdict
from pprint import pprint
import json
import logging
import os.path

import lxml.etree as ET

from .version import __version__
from . import dbhelper


_log = logging.getLogger(__name__)
#logging.basicConfig(
#    level=logging.INFO,
#    format='%(asctime)s %(levelname)-5s %(message)s',
#    datefmt='%Y-%m-%d %H:%M:%S',
#)

# Rename field in a dictionary
def ren(d, old, new):
    if old in d:
        d[new] = d[old]
        del d[old]


def as_bool(v):
    return {"false": 0, "true": 1}[v]


def dump_el(el):
    print(ET.tostring(el).decode())


class PortfolioPerformanceXML2DB:

    def parse_props(self, el, props):
        d = {}
        for p in props:
            conv = lambda x: x
            if isinstance(p, tuple):
                conv = p[1]
                p = p[0]
            pel = el.find(p)
            if pel is not None:
                d[p] = conv("" if pel.text is None else pel.text)
            elif p in el.attrib:
                # Otherwise try attribute (will return None if not there)
                d[p] = conv(el.get(p))
        return d

    def uuid(self, el):
        id = el.get("reference")
        if id is None:
            id = el.get("id")
        assert id is not None
        return self.id2uuid_map[id]

    @staticmethod
    def is_account_tag(tag):
        return tag in ("account", "referenceAccount", "accountFrom", "accountTo")

    def parse_entry(self, entry_el):
        els = entry_el.findall("*")
        assert len(els) == 2, len(els)
        return [(e.tag, e.text) for e in els]

    def parse_configuration(self, pel):
        conf = {}
        for c_el in pel.findall("configuration/entry"):
            d = self.parse_entry(c_el)
            assert d[0][0] == "string"
            assert d[1][0] == "string"
            conf[d[0][1]] = d[1][1] if d[1][1] is not None else ""
        return conf

    def parse_attributes(self, pel, el_tag="attributes/map"):
        attr_els = pel.findall(el_tag + "/entry")
        for seq, attr_el in enumerate(attr_els):
            els = attr_el.findall("*")
            assert len(els) == 2
            assert els[0].tag == "string"
            if els[1].tag == "limitPrice":
                fields = self.parse_props(els[1], ("operator", "value"))
                value = "%s %s" % (fields["operator"], fields["value"])
            elif els[1].tag == "bookmark":
                fields = self.parse_props(els[1], ("label", "pattern"))
                if not fields:
                    value = None
                else:
                    value = json.dumps(fields)
            else:
                value = els[1].text
            fields = {
                "attr_uuid": els[0].text,
                "type": els[1].tag,
                "value": value,
                "seq": seq,
            }
            yield fields

    def handle_price(self, price_el):
            props = ["t", "v"]
            price_fields = self.parse_props(price_el, props)
            ren(price_fields, "v", "value")
            ren(price_fields, "t", "tstamp")
            price_fields["security"] = self.cur_uuid()
            dbhelper.insert("price", price_fields)

    def handle_latest(self, latest_el):
        if latest_el is not None:
            props = ["t", "v", "high", "low", "volume"]
            latest_fields = self.parse_props(latest_el, props)
            ren(latest_fields, "v", "value")
            ren(latest_fields, "t", "tstamp")
            latest_fields["security"] = self.cur_uuid()
            dbhelper.insert("latest_price", latest_fields)

    def handle_event(self, event_el):
            props = ["date", "type", "details"]
            fields = self.parse_props(event_el, props)
            fields["security"] = self.cur_uuid()
            dbhelper.insert("security_event", fields)

    def handle_security(self, el):
        if el.get("reference") is not None:
            return

        props = [
            "uuid", "onlineId", "name", "currencyCode", "targetCurrencyCode", "note",
            "isin", "tickerSymbol", "calendar", "wkn", "feedTickerSymbol",
            "feed", "feedURL", "latestFeed", "latestFeedURL",
            ("isRetired", as_bool), "updatedAt"
        ]
        sec = self.parse_props(el, props)
        ren(sec, "currencyCode", "currency")
        ren(sec, "targetCurrencyCode", "targetCurrency")
        dbhelper.insert("security", sec)

        for fields in self.parse_attributes(el):
            fields["security"] = sec["uuid"]
            dbhelper.insert("security_attr", fields)

        prop_els = el.findall("property")
        for seq, prop_el in enumerate(prop_els):
            fields = {
                "security": sec["uuid"], "type": prop_el.get("type"),
                "name": prop_el.get("name"), "value": prop_el.text, "seq": seq,
            }
            dbhelper.insert("security_prop", fields)

    def handle_account_attrs(self, pel, uuid):
        for fields in self.parse_attributes(pel):
            fields["account"] = uuid
            dbhelper.insert("account_attr", fields)

    def handle_account(self, el, orderno):
        props = ["uuid", "name", "currencyCode", "note", ("isRetired", as_bool), "updatedAt", "id"]
        fields = self.parse_props(el, props)
        ren(fields, "currencyCode", "currency")
        ren(fields, "id", "_xmlid")
        fields["type"] = "account"
        fields["_order"] = orderno
        dbhelper.insert("account", fields)
        self.handle_account_attrs(el, fields["uuid"])

    def handle_portfolio(self, el, orderno):
        props = ["uuid", "name", "note", ("isRetired", as_bool), "updatedAt", "id"]
        fields = self.parse_props(el, props)
        ren(fields, "id", "_xmlid")
        acc = el.find("referenceAccount")
        fields["referenceAccount"] = self.uuid(acc)
        fields["type"] = "portfolio"
        fields["_order"] = orderno
        dbhelper.insert("account", fields)
        self.handle_account_attrs(el, fields["uuid"])

    def handle_watchlist(self, el):
        fields = self.parse_props(el, ["name"])
        id = dbhelper.insert("watchlist", fields, returning="_id")
        for sec in el.findall("securities/security"):
            fields = {"list": id, "security": self.uuid(sec)}
            dbhelper.insert("watchlist_security", fields)

    def handle_xact(self, acc_type, acc_uuid, el, orderno):
        # Start with calculating unit aggregates, to add to xact row in DB.
        units_dict = defaultdict(int)
        for unit_el in el.findall("units/unit"):
            am_el = unit_el.find("amount")
            units_dict[unit_el.get("type")] += int(am_el.get("amount"))

        props = [
            "uuid",
            "date",
            "currencyCode",
            "amount",
            "shares",
            "note",
            "source",
            "updatedAt",
            "type",
            "id",
        ]
        fields = self.parse_props(el, props)
        ren(fields, "currencyCode", "currency")
        ren(fields, "id", "_xmlid")
        fields["account"] = acc_uuid
        fields["acctype"] = acc_type
        fields["_order"] = orderno
        sec = el.find("security")
        if sec is not None:
            fields["security"] = self.uuid(sec)
        fields["fees"] = units_dict["FEE"]
        fields["taxes"] = units_dict["TAX"]
        dbhelper.insert("xact", fields)

        xact_uuid = fields["uuid"]
        for unit_el in el.findall("units/unit"):
            am_el = unit_el.find("amount")
            fields = {
                "xact": xact_uuid,
                "type": unit_el.get("type"),
                "amount": am_el.get("amount"),
                "currency": am_el.get("currency"),
            }
            forex_el = unit_el.find("forex")
            if forex_el is not None:
                fields["forex_amount"] = forex_el.get("amount")
                fields["forex_currency"] = forex_el.get("currency")
            rate_el = unit_el.find("exchangeRate")
            if rate_el is not None:
                fields["exchangeRate"] = rate_el.text
            dbhelper.insert("xact_unit", fields)

    def handle_crossEntry(self, x_el):
            if x_el.get("reference") is not None:
                return

            typ = x_el.get("class")
            if typ == "buysell":
                fields = {
                    "type": typ,
                    "from_acc": self.uuid(x_el.find("portfolio")),
                    "from_xact": self.uuid(x_el.find("portfolioTransaction")),
                    "to_acc": self.uuid(x_el.find("account")),
                    "to_xact": self.uuid(x_el.find("accountTransaction")),
                }
            elif typ == "account-transfer":
                fields = {
                    "type": typ,
                    "from_acc": self.uuid(x_el.find("accountFrom")),
                    "from_xact": self.uuid(x_el.find("transactionFrom")),
                    "to_acc": self.uuid(x_el.find("accountTo")),
                    "to_xact": self.uuid(x_el.find("transactionTo")),
                }
            elif typ == "portfolio-transfer":
                fields = {
                    "type": typ,
                    "from_acc": self.uuid(x_el.find("portfolioFrom")),
                    "from_xact": self.uuid(x_el.find("transactionFrom")),
                    "to_acc": self.uuid(x_el.find("portfolioTo")),
                    "to_xact": self.uuid(x_el.find("transactionTo")),
                }
            else:
                raise NotImplementedError(typ)
            dbhelper.insert("xact_cross_entry", fields)

    def handle_taxonomy(self, taxon_el):
            props = ["id", "name"]
            fields = self.parse_props(taxon_el, props)
            ren(fields, "id", "uuid")
            for dim_els in taxon_el.findall("dimensions/string"):
                dim_fields = {
                    "taxonomy": fields["uuid"],
                    "name": "dimension",
                    "value": dim_els.text,
                }
                dbhelper.insert("taxonomy_data", dim_fields)
            root_el = taxon_el.find("root")
            fields["root"] = self.uuid(root_el)
            dbhelper.insert("taxonomy", fields)
            self.handle_taxonomy_level(fields["uuid"], None, root_el)

    def handle_taxonomy_level(self, taxon_uuid, parent_uuid, level_el):
        props = ["id", "name", "color", "weight", "rank"]
        fields = self.parse_props(level_el, props)
        ren(fields, "id", "uuid")
        fields["parent"] = parent_uuid
        fields["taxonomy"] = taxon_uuid
        level_uuid = fields["uuid"]
        dbhelper.insert("taxonomy_category", fields)

        for data_el in level_el.findall("data/entry"):
            data = self.parse_entry(data_el)
            fields = {
                "name": data[0][1],
                "type": data[1][0],
                "value": data[1][1],
                "category": level_uuid,
                "taxonomy": taxon_uuid,
            }
            dbhelper.insert("taxonomy_data", fields)

        for as_el in level_el.findall("assignments/assignment"):
            props = ["weight", "rank"]
            fields = self.parse_props(as_el, props)
            el = as_el.find("investmentVehicle")
            fields["item_type"] = el.get("class")
            fields["item"] = self.uuid(el)
            fields["category"] = level_uuid
            fields["taxonomy"] = taxon_uuid
            id = dbhelper.insert("taxonomy_assignment", fields, returning="_id")
            for data_el in as_el.findall("data/entry"):
                data = self.parse_entry(data_el)
                fields = {
                    "assignment": id,
                    "name": data[0][1],
                    "type": data[1][0],
                    "value": data[1][1],
                }
                dbhelper.insert("taxonomy_assignment_data", fields)

        for ch_el in level_el.findall("children/classification"):
            self.handle_taxonomy_level(taxon_uuid, level_uuid, ch_el)

    def handle_dashboard(self, dashb_el):
            props = ["id", "name"]
            fields = self.parse_props(dashb_el, props)
            conf = self.parse_configuration(dashb_el)
            fields["config_json"] = json.dumps(conf)

            columns = []
            for col_el in dashb_el.findall("columns/column"):
                props = ["weight"]
                col_fields = self.parse_props(col_el, props)
                col_fields["widgets"] = []
                for widget_el in col_el.findall("widgets/widget"):
                    wid_fields = self.parse_props(widget_el, ["label"])
                    wid_fields["type"] = widget_el.get("type")
                    if widget_el.find("configuration") is not None:
                        conf = self.parse_configuration(widget_el)
                        wid_fields["config"] = conf
                    col_fields["widgets"].append(wid_fields)
                columns.append(col_fields)
            fields["columns_json"] = json.dumps(columns)
            dbhelper.insert("dashboard", fields)

    def handle_settings(self, settings_el):
        for bmark_el in settings_el.findall("bookmarks/bookmark"):
            props = ["label", "pattern"]
            fields = self.parse_props(bmark_el, props)
            dbhelper.insert("bookmark", fields)

        for attr_type_el in settings_el.findall("attributeTypes/attribute-type"):
            props = ["id", "name", "columnLabel", "source", "target", "type", "converterClass"]
            fields = self.parse_props(attr_type_el, props)
            props = []
            for p in self.parse_attributes(attr_type_el, "properties"):
                props.append({"name": p["attr_uuid"], "type": p["type"], "value": p["value"]})
            fields["props_json"] = json.dumps(props)
            dbhelper.insert("attribute_type", fields)

        for config_set_el in settings_el.findall("configurationSets/entry"):
            props = ["string"]
            fields = self.parse_props(config_set_el, props)
            ren(fields, "string", "name")
            cset_id = dbhelper.insert("config_set", fields, returning="_id")
            for config_e_el in config_set_el.findall("config-set/configurations/config"):
                props = ["uuid", "name", "data"]
                fields = self.parse_props(config_e_el, props)
                fields["config_set"] = cset_id
                dbhelper.insert("config_entry", fields)

    def handle_toplevel_properties(self, el):
        for prop_el in el.findall("entry"):
            d = self.parse_entry(prop_el)
            assert d[0][0] == "string"
            assert d[1][0] == "string"
            fields = {"name": d[0][1], "value": d[1][1]}
            dbhelper.insert("property", fields)

    def handle_client(self, el):
        props = ["version", "baseCurrency"]
        fields = self.parse_props(el, props)
        for n in props:
            dbhelper.insert("property", {"name": n, "value": fields[n], "special": 1})

    def __init__(self, xml, skip_prices=False):
        self.xml = xml
        self.refcache = {}
        self.skip_prices = skip_prices

    def parse(self):
        props = ["version", "baseCurrency"]
        fields = self.parse_props(self.etree, props)
        for n in props:
            dbhelper.insert("property", {"name": n, "value": fields[n], "special": 1})

        _log.info("Handling <security>")
        security_els = self.etree.findall("securities/security")
        for s in security_els:
            self.handle_security(s)

        _log.info("Handling <watchlist>")
        for w in self.etree.findall("watchlists/watchlist"):
            self.handle_watchlist(w)

        _log.info("Handling <account>")
        account_els = self.etree.findall("accounts/account")
        for el in account_els:
            self.handle_account(el)

        _log.info("Handling <portfolio>")
        portfolio_els = self.etree.findall("portfolios/portfolio")
        for el in portfolio_els:
            self.handle_portfolio(el)

        _log.info("Handling <account-transaction>")
        for acc_el in self.etree.xpath(".//*[self::account or self::accountTo]"):
            acc_uuid = self.uuid(acc_el)
            for xact_el in acc_el.findall(".//account-transaction"):
                self.handle_xact("account", acc_uuid, xact_el)

        _log.info("Handling <portfolio-transaction>")
        for acc_el in self.etree.xpath(".//*[self::portfolio or self::portfolioTo]"):
            acc_uuid = self.uuid(acc_el)
            for xact_el in acc_el.findall(".//portfolio-transaction"):
                self.handle_xact("portfolio", acc_uuid, xact_el)

        _log.info("Handling <crossEntry>")
        for x_el in self.etree.findall("//crossEntry"):
            self.handle_crossEntry(x_el)

        _log.info("Handling <taxonomy>")
        for taxon_el in self.etree.findall("taxonomies/taxonomy"):
            self.handle_taxonomy(taxon_el)

        _log.info("Handling <dashboard>")
        for dashb_el in self.etree.findall("dashboards/dashboard"):
            self.handle_dashboard(dashb_el)

        _log.info("Handling <properties>")
        self.handle_toplevel_properties(self.etree.find("properties"))

        _log.info("Handling <settings>")
        self.handle_settings(self.etree.find("settings"))

    def cur_uuid(self):
        return self.container_stack[-1][1]

    def iterparse(self):
        self.el_stack = []
        self.container_stack = []
        self.cur_xmlid = None
        self.id2uuid_map = {}
        self.uuid2ctr_map = {}
        self.el_order = 0
        for event, el in ET.iterparse(self.xml, events=("start", "end")):
            #print(event, el, el.attrib)
            self.el_order += 1
            if event == "start":
                self.el_stack.append(el.tag)
                if el.tag in ("security", "account", "referenceAccount", "accountFrom", "accountTo", "portfolio", "portfolioFrom", "portfolioTo"):
                    self.cur_xmlid = el.get("id")
                    if self.cur_xmlid is not None:
                        # Real element definition, not reference
                        self.container_stack.append([el.tag, None])
                        #print("Pushed on container stack:", self.container_stack)
                elif el.tag in ("account-transaction", "accountTransaction", "portfolio-transaction", "portfolioTransaction", "transactionFrom", "transactionTo"):
                    self.cur_xmlid = el.get("id")
                elif el.tag in ("root", "classification"):
                    self.cur_xmlid = el.get("id")
                elif el.tag in ("taxonomy", "dashboard", "settings"):
                    self.container_stack.append([el.tag, None])

            elif event == "end":
                assert self.el_stack[-1] == el.tag
                self.el_stack.pop()
                if el.tag in ("uuid", "id"):
                    if  self.container_stack and self.container_stack[-1][1] is None:
                        self.container_stack[-1][1] = el.text
                        #print("Setting uuid of top container:", self.container_stack, el.sourceline)
                        self.uuid2ctr_map[el.text] = self.container_stack[-1][0]
                    self.id2uuid_map[self.cur_xmlid] = el.text

                elif el.tag == "price":
                    if not self.skip_prices:
                        self.handle_price(el)
                elif el.tag == "latest":
                    self.handle_latest(el)
                elif el.tag == "event":
                    self.handle_event(el)

                elif el.tag == "security":
                    self.handle_security(el)
                elif el.tag == "watchlist":
                    self.handle_watchlist(el)
                elif el.tag in ("account", "accountFrom", "accountTo", "referenceAccount"):
                    if el.get("id"):
                        self.handle_account(el, self.el_order)
                    elif el.tag == "account":
                        xmlid = el.get("reference")
                        dbhelper.execute_dml("UPDATE account SET _order=? WHERE _xmlid=?", (self.el_order, xmlid))
                elif el.tag in ("portfolio", "portfolioFrom", "portfolioTo"):
                    if el.get("id"):
                        self.handle_portfolio(el, self.el_order)
                    elif el.tag == "portfolio":
                        xmlid = el.get("reference")
                        dbhelper.execute_dml("UPDATE account SET _order=? WHERE _xmlid=?", (self.el_order, xmlid))

                elif el.tag == "account-transaction":
                    if el.get("id"):
                        assert self.is_account_tag(self.uuid2ctr_map[self.cur_uuid()]), self.uuid2ctr_map[self.cur_uuid()]
                        self.handle_xact("account", self.cur_uuid(), el, self.el_order)
                    else:
                        xmlid = el.get("reference")
                        dbhelper.execute_dml("UPDATE xact SET _order=? WHERE _xmlid=?", (self.el_order, xmlid))

                elif el.tag == "accountTransaction":
                    if el.get("id"):
                        parent = el.getparent()
                        uuid = self.uuid(parent.find("account"))
                        assert self.is_account_tag(self.uuid2ctr_map[uuid]), self.uuid2ctr_map[uuid]
                        self.handle_xact("account", uuid, el, 0)

                elif el.tag in ("portfolio-transaction",):
                    if el.get("id"):
                        assert self.uuid2ctr_map[self.cur_uuid()].startswith("portfolio")
                        self.handle_xact("portfolio", self.cur_uuid(), el, self.el_order)
                    else:
                        xmlid = el.get("reference")
                        dbhelper.execute_dml("UPDATE xact SET _order=? WHERE _xmlid=?", (self.el_order, xmlid))

                elif el.tag in ("portfolioTransaction",):
                    if el.get("id"):
                        parent = el.getparent()
                        uuid = self.uuid(parent.find("portfolio"))
                        assert self.uuid2ctr_map[uuid].startswith("portfolio"), self.uuid2ctr_map[uuid]
                        self.handle_xact("portfolio", uuid, el, 0)

                elif el.tag == "transactionTo":
                    if el.get("id"):
                        parent = el.getparent()
                        assert parent.tag == "crossEntry"
                        if parent.get("class") == "account-transfer":
                            what = "account"
                            uuid = self.uuid(parent.find("accountTo"))
                        elif parent.get("class") == "portfolio-transfer":
                            what = "portfolio"
                            uuid = self.uuid(parent.find("portfolioTo"))
                        else:
                            assert False, "Unexpected crossEntry class: " + parent.get("class")

                        if what == "account":
                            assert self.is_account_tag(self.uuid2ctr_map[uuid]), self.uuid2ctr_map[uuid]
                        else:
                            assert self.uuid2ctr_map[uuid].startswith(what), self.uuid2ctr_map[uuid]
                        self.handle_xact(what, uuid, el, 0)
                elif el.tag == "transactionFrom":
                    if el.get("id"):
                        parent = el.getparent()
                        assert parent.tag == "crossEntry"
                        if parent.get("class") == "account-transfer":
                            what = "account"
                            uuid = self.uuid(parent.find("accountFrom"))
                        elif parent.get("class") == "portfolio-transfer":
                            what = "portfolio"
                            uuid = self.uuid(parent.find("portfolioFrom"))
                        else:
                            assert False, "Unexpected crossEntry class: " + parent.get("class")

                        if what == "account":
                            assert self.is_account_tag(self.uuid2ctr_map[uuid]), self.uuid2ctr_map[uuid]
                        else:
                            assert self.uuid2ctr_map[uuid].startswith(what), self.uuid2ctr_map[uuid]
                        self.handle_xact(what, uuid, el, 0)

                elif el.tag == "crossEntry":
                    if el.get("id"):
                        self.handle_crossEntry(el)

                elif el.tag == "taxonomy":
                    self.handle_taxonomy(el)
                elif el.tag == "dashboard":
                    self.handle_dashboard(el)
                elif el.tag == "settings":
                    self.handle_settings(el)
                elif el.tag == "properties" and self.el_stack[-1] == "client":
                    self.handle_toplevel_properties(el)
                elif el.tag == "client":
                    self.handle_client(el)

                if el.get("reference") is None and self.container_stack and self.container_stack[-1][0] == el.tag:
                    self.container_stack.pop()

                # To save memory, we clear children of processed elements,
                # execept for cases below.
                preserve = False
                if self.container_stack and self.container_stack[-1][0] in ("taxonomy", "dashboard", "settings"):
                    preserve = True
                elif el.tag in ("units", "unit"):
                    preserve = True
                elif el.tag in ("limitPrice", "bookmark"):
                    preserve = True
                elif el.tag in ("map", "entry"):
                    preserve = True
                elif self.el_stack and self.el_stack[-1] == "watchlist" and el.tag == "securities":
                    preserve = True
                elif self.container_stack and self.container_stack[-1][0] in ("security", "account", "portfolio", "accountFrom", "accountTo", "referenceAccount", "portfolioFrom", "portfolioTo") and el.tag == "attributes":
                    preserve = True

                if not preserve:
                    # Remove children and text of elements. We don't use
                    # el.clear(), as that also removed attributes, but
                    # we want to preserve them (need id/reference at least).
                    for ch in list(el):
                        el.remove(ch)
                        el.text = el.tail = None


if __name__ == "__main__":
    argp = argparse.ArgumentParser(description="Import PortfolioPerformance XML file to Sqlite DB")
    argp.add_argument("xml_file", help="input XML file")
    argp.add_argument("db", help="output DB (filename/connect string)")
    argp.add_argument("--dbtype", choices=("sqlite", "pgsql"), default="sqlite", help="select database type")
    argp.add_argument("--debug", action="store_true", help="enable debug logging")
    argp.add_argument("--dry-run", action="store_true", help="don't commit changes to DB")
    argp.add_argument("--skip-prices", action="store_true", help="don't import historical prices (95+%% of DB size and import time; useful for debugging)")
    argp.add_argument("--version", action="version", version="%(prog)s " + __version__)
    args = argp.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    dbhelper.init(args.dbtype, args.db)

    with open(args.xml_file, "rb") as f:
        conv = PortfolioPerformanceXML2DB(f)
        conv.iterparse()

    if not args.dry_run:
        dbhelper.commit()



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from typing import Any, Callable, Dict, Optional

import numpy as np
import pandas as pd

log = logging.getLogger(__name__)

def _percent_plain_converter(value: Any) -> float:
    """Convert percent value stored as plain number (divide by 100)."""
    return float(value) / 100


CONVERTER_DISPATCH: Dict[str, Callable[[Any], Any]] = {
    'DateConverter': pd.to_datetime,
    'PercentPlainConverter': _percent_plain_converter,
    'PercentConverter': float,
    'LongConverter': float,
    'AmountConverter': float,
    'StringConverter': str,
}


def _get_converter_function(converter: str) -> Optional[Callable[[Any], Any]]:
    """Get the conversion function for a given converter class name."""
    for key, func in CONVERTER_DISPATCH.items():
        if key in converter:
            return func
    return None


def _convert_single_value(value: Any, converter: Any, attr_name: str, attr_uuid: str, idx: Any) -> Any:
    """
    Convert a single attribute value based on its converter type.

    Args:
        value: The raw attribute value to convert
        converter: The converter class name
        attr_name: Friendly name of the attribute (for logging)
        attr_uuid: UUID of the attribute (for logging)
        idx: Row index (for logging)

    Returns:
        Converted value or np.nan if conversion fails
    """
    if pd.isna(value):
        return value

    if pd.isna(converter):
        log.warning(
            "Missing converter type for attribute '%s' (%s) of entity at index %s. Ignoring value.",
            attr_name, attr_uuid, idx
        )
        return np.nan

    converter_str = str(converter)
    convert_func = _get_converter_function(converter_str)

    if convert_func is None:
        log.debug(
            "Unknown converter type '%s' for attribute '%s' (%s). Keeping raw value.",
            converter_str, attr_name, attr_uuid
        )
        return value

    try:
        return convert_func(value)
    except (ValueError, TypeError) as e:
        log.warning(
            "Failed to parse attribute '%s' (%s) value '%s': %s. Ignoring value.",
            attr_name, attr_uuid, value, str(e)
        )
        return np.nan


def convert_attribute_types(df: pd.DataFrame, attributes: Dict[str, str]) -> pd.DataFrame:
    """
    Convert attribute values based on their converterClass types.

    Portfolio Performance stores attribute metadata including a converterClass that indicates
    how to interpret the stored string value. This function transforms those string values
    into their proper Python types (float, datetime, etc.).

    Args:
        df: DataFrame containing attribute columns (as UUIDs) and corresponding
            {uuid}_converter columns with converter class names
        attributes: Dictionary mapping attribute UUIDs to their friendly names

    Returns:
        DataFrame with converted attribute values and converter columns removed
    """
    for attr_uuid, attr_name in attributes.items():
        value_col = attr_uuid
        converter_col = get_converter_column_name(attr_uuid)

        if value_col not in df.columns:
            continue

        df[value_col] = df.apply(
            lambda row, value_column=value_col, converter_column=converter_col, attribute_name=attr_name, attribute_uuid=attr_uuid: _convert_single_value(
                row.get(value_column),
                row.get(converter_column),
                attribute_name,
                attribute_uuid,
                row.name
            ),
            axis=1
        )

        if converter_col in df.columns:
            df = df.drop(columns=[converter_col])

    return df

def get_converter_column_name(attr_uuid: str) -> str:
    return f"_{attr_uuid}_converter"



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from pathlib import Path
import random
from datetime import datetime, timedelta
import logging
from typing import Any

import lxml.etree as ET  # pylint: disable=c-extension-no-member
from faker import Faker

log = logging.getLogger(__name__)


class XmlAnonymizer:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """Anonymizes Portfolio Performance XML files."""

    # Tags that should not be anonymized
    SKIP_TAGS = {
        'uuid', 'isin', 'wkn', 'tickerSymbol', 'feed', 'feedURL',
        'currencyCode', 'targetCurrencyCode', 'calendar', 'type',
        'isRetired', 'updatedAt', 'version', 'baseCurrency'
    }

    def __init__(
        self,
        seed: int,
        config: dict[str, Any] | None = None,
        date_shift_days_range: tuple[int, int] = (-3650, -365),
        amount_factor_range: tuple[float, float] = (0.5, 2.0)
    ):
        self.seed = seed
        self.date_shift_days_range = date_shift_days_range
        self.amount_factor_range = amount_factor_range

        self.rng = random.Random(seed)  # nosec B311  # Used for data anonymization, not cryptography
        self.faker = Faker()
        Faker.seed(seed)

        self.config = config if config is not None else {}

        # Caches for consistent anonymization
        self.amount_factors: dict[str, float] = {}
        self.date_offset: int | None = None

    def anonymize_file(self, input_path: Path, output_path: Path) -> None:
        parser = ET.XMLParser(remove_blank_text=False, resolve_entities=False, no_network=True)  # pylint: disable=c-extension-no-member
        tree = ET.parse(str(input_path), parser)  # pylint: disable=c-extension-no-member
        root = tree.getroot()

        self._init_date_offset()
        self._anonymize_element(root)

        tree.write(
            str(output_path),
            encoding='UTF-8',
            xml_declaration=True,
            pretty_print=False
        )

    def _init_date_offset(self) -> None:
        """Initialize random date offset."""
        min_days, max_days = self.date_shift_days_range
        self.date_offset = self.rng.randint(min_days, max_days)

    def _anonymize_element(self, element: ET.Element) -> None:  # pylint: disable=too-many-branches,c-extension-no-member
        """Recursively anonymize XML element and children."""
        tag = element.tag

        if tag in self.SKIP_TAGS:
            # Still process children
            for child in element:
                self._anonymize_element(child)
            return

        if tag == 'name':
            element.text = self._generate_name(element)
        elif tag in ('note', 'details'):
            element.text = self.faker.sentence() if element.text else None
        elif tag == 'date':
            element.text = self._shift_date(element.text)
        elif tag == 'amount':
            element.text = self._randomize_amount(element.text, element)
            # Also handle amount attribute for multi-currency amounts
            if 'amount' in element.attrib:
                element.attrib['amount'] = self._randomize_amount(
                    element.attrib['amount'], element
                )
        elif tag == 'shares':
            element.text = self._randomize_amount(element.text, element)
        elif tag == 'price':
            # Price element has 't' (date) and 'v' (value) attributes
            if 't' in element.attrib:
                element.attrib['t'] = self._shift_date(element.attrib['t'])
            if 'v' in element.attrib:
                element.attrib['v'] = self._randomize_amount(
                    element.attrib['v'], element
                )
        elif tag == 'source':
            element.text = self.faker.file_name(extension="pdf") if element.text else None
        elif tag == 'entry' and element.getparent() is not None and element.getparent().tag == 'map':
            self._anonymize_attribute_entry(element)

        for child in element:
            self._anonymize_element(child)

    def _generate_name(self, element: ET.Element) -> str:  # pylint: disable=c-extension-no-member
        if not element.text or not element.text.strip():
            return ''

        parent = element.getparent()
        if parent is None:
            return self.faker.company()

        parent_tag = parent.tag
        if 'account' in parent_tag.lower():
            return f"{self.faker.word().capitalize()} Account"
        if 'portfolio' in parent_tag.lower():
            return f"{self.faker.word().capitalize()} Portfolio"

        return str(element.text)

    def _shift_date(self, date_str: str | None) -> str:
        """Shift date by random offset."""
        if not date_str or not self.date_offset:
            return ''

        try:
            if 'T' in date_str:
                # ISO format with time: "2019-01-01T00:00"
                # Handle optional Z suffix
                clean_date = date_str.replace('Z', '+00:00')
                dt = datetime.fromisoformat(clean_date)
                shifted = dt + timedelta(days=self.date_offset)
                # Preserve original format
                if date_str.endswith('Z'):
                    return shifted.isoformat().replace('+00:00', 'Z')
                return shifted.isoformat()

            # Date only: "2019-01-01"
            dt = datetime.strptime(date_str, '%Y-%m-%d')
            shifted = dt + timedelta(days=self.date_offset)
            return shifted.strftime('%Y-%m-%d')
        except (ValueError, AttributeError):
            return date_str

    def _randomize_amount(self, amount_str: str | None, element: ET.Element) -> str:  # pylint: disable=c-extension-no-member
        """Randomize amount while preserving order of magnitude."""
        if not amount_str:
            return ''

        try:
            # Build context key for consistency
            parent = element.getparent()
            context = f"{element.tag}_{parent.tag if parent is not None else ''}"

            # Get or generate factor
            if context not in self.amount_factors:
                min_f, max_f = self.amount_factor_range
                self.amount_factors[context] = self.rng.uniform(min_f, max_f)

            factor = self.amount_factors[context]
            value = int(amount_str)
            new_value = int(value * factor)

            return str(new_value)
        except (ValueError, AttributeError):
            return amount_str

    def _anonymize_attribute_entry(self, entry: ET.Element) -> None:  # pylint: disable=c-extension-no-member
        strings = entry.findall('string')
        if len(strings) < 2:
            return

        attr_type_uuid = strings[0].text
        current_value = strings[1].text

        if not attr_type_uuid or not current_value:
            return

        if attr_type_uuid in self.config:
            config = self.config[attr_type_uuid]
            provider = config.get('provider')
            args = config.get('args', {})

            if provider:
                new_value = self._call_faker_provider(provider, args)
                if new_value is not None:
                    strings[1].text = new_value

    def _call_faker_provider(self, provider: str, args: dict[str, Any]) -> str | None:
        method = getattr(self.faker, provider)
        return str(method(**args))



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from pp_terminal.exceptions import InputError


def normalize_columns(
    requested_columns: list[str],
    available_columns: list[str],
    attributes: dict[str, str] | None = None
) -> list[str]:
    """
    Normalize and validate column names (case-insensitive matching).

    Args:
        requested_columns: List of column names requested by user (UUIDs or regular column names)
        available_columns: List of actual column names in the dataframe
        attributes: Optional mapping of UUID to friendly attribute names (for error messages only)

    Returns:
        List of normalized column names matching the dataframe columns
    """
    normalized = []
    available_lower = {col.lower(): col for col in available_columns}

    if 'currency' in available_columns:
        normalized.append(available_lower['currency'])

    for col in requested_columns:
        col_lower = col.strip().lower()

        if col_lower in available_lower:
            normalized.append(available_lower[col_lower])
        else:
            # Build helpful error message with both UUIDs and friendly names
            uuid_keys = set(attributes.keys()) if attributes else set()
            available_names = sorted([
                col for col in available_columns
                if not col.startswith('_') and col not in uuid_keys
            ])
            # Add attribute columns with their friendly names
            if attributes:
                for uuid, name in sorted(attributes.items(), key=lambda x: x[1]):
                    available_names.append(f"{uuid} ({name})")
            raise InputError(f"Column '{col}' not found. Available columns: {', '.join(available_names)}")

    return list(dict.fromkeys(normalized))  # make unique



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Any, cast
import logging
import pandas as pd
from pandera.typing import DataFrame

from pp_terminal.data.cost_basis import calculate_total_cost_basis
from pp_terminal.data.filters import filter_by_security
from pp_terminal.data.tax import load_prepaid_tax_data_from_csv
from pp_terminal.domain.portfolio import Portfolio
from pp_terminal.domain.schemas import TaxPaidSchema

log = logging.getLogger(__name__)


class ValidationRule(ABC):
    def __init__(
        self,
        rule_type: str,
        value: Any,
        severity: str = 'error',
        applies_to: list[str] | None = None
    ):
        self.rule_type = rule_type
        self._value = value
        self.severity = severity
        self.applies_to = applies_to

    @abstractmethod
    def validate(self, entity: pd.Series, entity_id: str, context: dict[str, Any]) -> tuple[bool, str | None]:
        """
        Validate entity and return (is_error, message) tuple.

        Returns:
            tuple[bool, str | None]:
                - First element: True if error occurred (severity='error' and validation failed)
                - Second element: Violation message if validation failed, None otherwise
        """
        log.debug('Validating %s of "%s" (%s) using value %s %s', str(self), entity["name"], entity_id, str(self._get_value(entity)), '(' + str(self._value) + ')' if self._value != self._get_value(entity) else '')

        return (False, None)

    def matches_entity(self, entity: pd.Series, entity_id: str) -> bool:
        if self.rule_type.endswith('-from-attribute'):
            attr_uuid = self._value
            return attr_uuid in entity.index and pd.notna(entity.get(attr_uuid))

        if self.applies_to is not None:
            return entity_id in self.applies_to

        return True

    def _get_value(self, entity: pd.Series) -> Any:
        if self.rule_type.endswith('-from-attribute'):
            attr_uuid = self._value
            return entity.get(attr_uuid)
        return self._value

    def log_violation(self, message: str) -> None:
        if self.severity == 'error':
            log.error(message)
        else:
            log.warning(message)

    def is_error(self) -> bool:
        return bool(self.severity == 'error')

    def __str__(self) -> str:
        return self.rule_type


class BalanceLimitRule(ValidationRule):
    def validate(self, entity: pd.Series, entity_id: str, context: dict[str, Any]) -> tuple[bool, str | None]:
        super().validate(entity, entity_id, context)

        limit = self._get_value(entity)
        balance = context['balance']

        if balance > limit:
            message = f'balance {balance:.2f} exceeds limit {limit:.2f}'
            return self.is_error(), message
        return False, None


class DatePassedRule(ValidationRule):
    def validate(self, entity: pd.Series, entity_id: str, context: dict[str, Any]) -> tuple[bool, str | None]:
        super().validate(entity, entity_id, context)

        date_value = self._get_value(entity)

        if pd.isna(date_value):
            return False, None

        if not isinstance(date_value, datetime):
            try:
                date_value = pd.to_datetime(date_value)
            except (ValueError, TypeError):
                log.warning('"%s" has invalid date value: %s', entity["name"], date_value)
                return False, None

        attribute_name = 'date attribute'
        portfolio = cast(Portfolio, context.get('portfolio')) if context else None
        if portfolio is not None:
            attribute_name = portfolio.all_attributes.get(self._value, attribute_name)

        current_date = datetime.now()
        if date_value < current_date:
            message = f'{attribute_name} has passed {date_value.strftime("%Y-%m-%d")}'
            return self.is_error(), message
        return False, None


class PriceStalenessRule(ValidationRule):
    def validate(self, entity: pd.Series, entity_id: str, context: dict[str, Any]) -> tuple[bool, str | None]:
        super().validate(entity, entity_id, context)

        max_days = self._get_value(entity)
        latest_price_date = context.get('latest_price_date')

        if pd.isna(latest_price_date) or latest_price_date is None:
            message = 'no price data'
            return self.is_error(), message

        current_date = datetime.now()
        days_old = (current_date - latest_price_date).days

        if days_old > max_days:
            message = f'price is {days_old} days old (latest: {latest_price_date.strftime("%Y-%m-%d")})'
            return self.is_error(), message
        return False, None


class PriceLimitRule(ValidationRule):
    def validate(self, entity: pd.Series, entity_id: str, context: dict[str, Any]) -> tuple[bool, str | None]:
        super().validate(entity, entity_id, context)

        limit = self._get_value(entity)
        current_price = context.get('current_price')

        if pd.isna(current_price):
            message = 'no price data'
            return self.is_error(), message

        if current_price >= limit:
            message = f'price {current_price:.2f} has reached limit {limit:.2f}'
            return self.is_error(), message
        return False, None


class PurchaseCostLimitRule(ValidationRule):
    def validate(self, entity: pd.Series, entity_id: str, context: dict[str, Any]) -> tuple[bool, str | None]:
        super().validate(entity, entity_id, context)

        limit = self._get_value(entity)
        portfolio = cast(Portfolio, context.get('portfolio'))

        if portfolio is None:
            raise RuntimeError('No portfolio in context for purchase-cost-limit validation')

        current_cost = calculate_total_cost_basis(portfolio.securities_account_transactions.pipe(filter_by_security, security_id=entity_id))

        if current_cost > limit:
            currency = entity.get('currency', 'EUR')
            message = f'current cost basis {current_cost:.2f} {currency} exceeds limit {limit:.2f} {currency}'
            return self.is_error(), message

        return False, None

    @staticmethod
    def _load_tax_csv(context: dict[str, Any]) -> DataFrame[TaxPaidSchema] | None:
        """Load tax credit CSV from config if available."""
        config = context.get('config', {})
        tax_csv_path = config.get('tax', {}).get('file')
        tax_rate = config.get('tax', {}).get('rate', 0)

        if not tax_csv_path:
            return None

        return load_prepaid_tax_data_from_csv(Path(tax_csv_path), tax_rate)


RULE_TYPES = {
    'balance-limit': BalanceLimitRule,
    'balance-limit-from-attribute': BalanceLimitRule,
    'date-passed-from-attribute': DatePassedRule,
    'price-staleness': PriceStalenessRule,
    'price-limit': PriceLimitRule,
    'price-limit-from-attribute': PriceLimitRule,
    'purchase-cost-limit': PurchaseCostLimitRule,
    'purchase-cost-limit-from-attribute': PurchaseCostLimitRule,
}


def create_rule(rule_config: dict[str, Any]) -> ValidationRule:
    rule_type = rule_config['type']
    if rule_type not in RULE_TYPES:
        raise ValueError(f'Unknown rule type: {rule_type}')

    value = rule_config['value']
    severity = rule_config.get('severity', 'error')
    applies_to = rule_config.get('applies-to', None)

    rule_class = RULE_TYPES[rule_type]
    return rule_class(
        rule_type=rule_type,
        value=value,
        severity=severity,
        applies_to=applies_to
    )  # type: ignore[abstract]


def get_applicable_rules(entity_id: str, entity: pd.Series, rules: list[ValidationRule]) -> list[ValidationRule]:
    applicable_rules: list[ValidationRule] = []

    for rule in rules:
        if rule.matches_entity(entity, entity_id):
            applicable_rules.append(rule)

    return applicable_rules



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from datetime import datetime
from functools import wraps
from typing import Callable, Any, cast

import typer
from typer.models import CommandFunctionType

from pp_terminal.domain.portfolio import Portfolio, get_security_by_id
from pp_terminal.exceptions import ValidationError
from pp_terminal.utils.config import Config
from pp_terminal.utils.helper import run_all_group_cmds
from pp_terminal.output.strategy import Console
from pp_terminal.domain.portfolio_snapshot import PortfolioSnapshot
from pp_terminal.validation.engine import validate_accounts
from pp_terminal.validation.engine import validate_securities

app = typer.Typer()
console = Console()
log = logging.getLogger(__name__)

validate_app = typer.Typer()
app.add_typer(validate_app, name="validate", help='Run configured validation rules on the portfolio data.')

exit_code = 0  # pylint: disable=invalid-name


def catch_errors(func: CommandFunctionType) -> Callable[..., CommandFunctionType]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        global exit_code  # pylint: disable=global-statement

        try:
            return func(*args, **kwargs)
        except ValidationError:
            exit_code = 1
            ctx = kwargs['ctx'] if 'ctx' in kwargs else args[0] if len(args) > 0 else None

            if ctx and ctx.invoked_subcommand is None:
                raise typer.Exit(exit_code)  # pylint: disable=raise-missing-from

            return None

    return wrapper


@validate_app.command(name="securities")
@catch_errors
def log_validate_securities(ctx: typer.Context) -> None:
    """Validate security prices."""
    portfolio = cast(Portfolio, ctx.obj.portfolio)
    config = cast(Config, ctx.obj.config)

    results = validate_securities(portfolio, config)

    if not results:
        log.debug('No security validation rules configured or no securities to validate')
        return

    has_errors = False
    for security_id, result in results.items():
        if not result.violations:
            continue

        security = get_security_by_id(portfolio, security_id)

        for rule, message in result.violations:
            full_message = f'Security "{security.name}" ({security_id}) {message}'
            rule.log_violation(full_message)
            if rule.is_error():
                has_errors = True

    if has_errors:
        raise ValidationError()


@validate_app.command(name="accounts")
@catch_errors
def log_validate_accounts(ctx: typer.Context) -> None:
    """Validate deposit accounts using configured validation rules."""
    portfolio = cast(Portfolio, ctx.obj.portfolio)
    config = cast(Config, ctx.obj.config)

    snapshot = PortfolioSnapshot(portfolio, datetime.now())
    results = validate_accounts(portfolio, snapshot, config)

    if not results:
        log.debug('No account validation rules configured or no accounts to validate')
        return

    has_errors = False
    for account_id, result in results.items():
        if not result.violations:
            continue

        account_name = portfolio.deposit_accounts.loc[account_id, 'name']

        for rule, message in result.violations:
            full_message = f'Account "{account_name}" ({account_id}) {message}'
            rule.log_violation(full_message)
            if rule.is_error():
                has_errors = True

    if has_errors:
        raise ValidationError()


@validate_app.callback(invoke_without_command=True)
@run_all_group_cmds(validate_app)
def validate_all(ctx: typer.Context) -> None:  # pylint: disable=unused-argument
    if ctx.invoked_subcommand is None:
        raise typer.Exit(exit_code)



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from pathlib import Path
from typing import Any
import uuid

import pandas as pd
from openpyxl.styles import Font, PatternFill, numbers
from openpyxl.utils import get_column_letter

from .strategy import OutputStrategy
from .table_decorator import TableOptions, camel_case_to_title
from ..domain.schemas import Money

_HEADER_BG_COLOR = 'D3D3D3'

class ExcelOutputStrategy(OutputStrategy):
    def __init__(self) -> None:
        self._filename: Path | None = None
        self._sheet_counter = 0

    def result_table(self, df: pd.DataFrame | None, options: TableOptions) -> Any:  # pylint: disable=too-many-locals
        if df is None:
            return self.empty_result()

        # Generate filename on first call
        if self._filename is None:
            unique_id = uuid.uuid4().hex[:8]
            self._filename = Path.cwd() / f"output_{unique_id}.xlsx"

        self._sheet_counter += 1
        sheet_name = self._sanitize_sheet_name(options.title) if options.title else f"Sheet{self._sheet_counter}"

        formatted_df = self._prepare_dataframe(df, options)

        mode = 'a' if self._filename.exists() else 'w'
        if_sheet_exists = 'replace' if mode == 'a' else None

        with pd.ExcelWriter(
            self._filename,
            engine='openpyxl',
            mode=mode,
            if_sheet_exists=if_sheet_exists
        ) as writer:
            formatted_df.to_excel(
                writer,
                sheet_name=sheet_name,
                index=options.show_index
            )

            worksheet = writer.sheets[sheet_name]
            self._apply_formatting(worksheet, formatted_df, options, df)

        if self._sheet_counter == 1:
            return (f"Created Excel file: {self._filename}\n", )
        return None

    @staticmethod
    def _prepare_dataframe(df: pd.DataFrame, options: TableOptions) -> pd.DataFrame:
        """Prepare DataFrame by adding totals row if needed."""
        formatted_df = df.copy()

        if options.show_total:
            summary_row = df.select_dtypes(include='number').sum()
            if 'currency' in df.columns:
                summary_row['currency'] = df['currency'].mode()[0] if not df['currency'].mode().empty else ''

            # Add 'Total' label to first non-currency column
            first_col = next((col for col in df.columns if col != 'currency'), None)
            if first_col:
                summary_row[first_col] = 'Total'

            formatted_df = pd.concat([formatted_df, summary_row.to_frame().T], ignore_index=True)

        if 'currency' in formatted_df.columns:
            formatted_df = formatted_df.drop(columns=['currency'])

        return formatted_df

    @staticmethod
    def _apply_formatting(worksheet: Any, formatted_df: pd.DataFrame, options: TableOptions, original_df: pd.DataFrame) -> None:  # pylint: disable=too-many-locals
        """Apply rich formatting to the worksheet."""
        # Format header row
        gray_fill = PatternFill(start_color=_HEADER_BG_COLOR, end_color=_HEADER_BG_COLOR, fill_type='solid')
        for idx, cell in enumerate(worksheet[1]):
            cell.font = Font(bold=True)
            cell.fill = gray_fill
            # Apply camel_case_to_title to column headers (skip index column if present)
            if options.show_index and idx == 0:
                continue  # Keep index header as-is
            col_idx = idx - 1 if options.show_index else idx
            if col_idx < len(formatted_df.columns):
                cell.value = camel_case_to_title(str(formatted_df.columns[col_idx]))

        # Freeze top row
        worksheet.freeze_panes = 'A2'

        # Auto-adjust column widths and apply number formats
        for idx, column in enumerate(formatted_df.columns, start=1):
            column_letter = get_column_letter(idx)
            max_length = max(
                len(str(column)),
                formatted_df[column].astype(str).str.len().max()
            )
            worksheet.column_dimensions[column_letter].width = min(max_length + 2, 50)

            # Apply currency formatting to Money columns
            if column in original_df.columns and len(original_df) > 0:
                sample_value = original_df[column].iloc[0]
                if isinstance(sample_value, Money):
                    # Apply currency format to data cells (skip header)
                    for row_idx in range(2, len(formatted_df) + 2):
                        cell = worksheet[f"{column_letter}{row_idx}"]
                        cell.number_format = numbers.FORMAT_CURRENCY_EUR_SIMPLE

        # Bold totals row if present
        if options.show_total and len(formatted_df) > 0:
            total_row_idx = len(formatted_df) + 1
            for cell in worksheet[total_row_idx]:
                cell.font = Font(bold=True)

    @staticmethod
    def _sanitize_sheet_name(name: str) -> str:
        """Sanitize sheet name for Excel compatibility."""
        invalid_chars = [':', '\\', '/', '?', '*', '[', ']']
        sanitized = name
        for char in invalid_chars:
            sanitized = sanitized.replace(char, '')
        # Truncate to 31 characters (Excel limit)
        return sanitized[:31]



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import hashlib
import logging
from pathlib import Path
from string import Template

log = logging.getLogger(__name__)

_FILE_SUFFIX = ".pp-terminal.db"


def _calculate_xml_checksum(xml_path: Path) -> str:
    """
    Calculate SHA-256 checksum of XML file.

    Reads file in binary chunks to handle large files efficiently.

    Args:
        xml_path: Path to XML file

    Returns:
        SHA-256 hex digest string (64 characters)

    Raises:
        OSError: If file cannot be read
    """
    sha256 = hashlib.sha256()
    with open(xml_path, 'rb') as f:
        while chunk := f.read(8192):
            sha256.update(chunk)
    return sha256.hexdigest()


def _get_cache_filename_template(xml_file: Path) -> Template:
    prefix = xml_file.stem if xml_file.name.startswith(".") else f".{xml_file.stem}"
    return Template(f"{prefix}.$checksum{_FILE_SUFFIX}")


def get_cache_path(xml_file: Path) -> Path:
    checksum = _calculate_xml_checksum(xml_file)
    return xml_file.parent / _get_cache_filename_template(xml_file).substitute(checksum=checksum)


def cleanup_old_cache_files(xml_file: Path) -> None:
    """
    Delete old cache files with different checksums.

    Searches for cache files matching pattern: [.]<xml-filename>.*{_FILE_SUFFIX}
    Deletes all except the one with current checksum.

    Args:
        xml_file: Path to XML file (original or anonymized)

    Raises:
        OSError: If cache files cannot be deleted (logged as warning, not raised)
    """

    cache_pattern = _get_cache_filename_template(xml_file).substitute(checksum="*")
    current_cache_name = get_cache_path(xml_file).name

    try:
        for cache_file in xml_file.parent.glob(cache_pattern):
            if cache_file.name == current_cache_name:
                continue

            try:
                cache_file.unlink()
                log.debug('Deleted old cache file "%s"', cache_file)
            except FileNotFoundError:
                # File already deleted (race condition), ignore
                pass
            except OSError as e:
                log.warning('Failed to delete old cache file "%s": %s', cache_file, str(e))
    except Exception as e:  # pylint: disable=broad-exception-caught
        log.warning('Failed to cleanup old cache files: %s', str(e))



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import json
import logging
from pathlib import Path
from typing import Any, Dict, cast

from jsonschema import Draft7Validator, ValidationError as JsonSchemaValidationError
from typer_config import conf_callback_factory
from typer_config.loaders import toml_loader

log = logging.getLogger(__name__)
type Config = Dict[str, Any]

# Global storage for the loaded config
_loaded_config: Config = {}



def _load_schema() -> dict[str, Any]:
    schema_path = Path(__file__).parent.parent / 'config.schema.json'
    with open(schema_path, 'r', encoding='utf-8') as f:
        return cast(dict[str, Any], json.load(f))


def validated_toml_loader(config_path: str) -> Config:
    """
    Load and validate TOML configuration file for use with typer-config.

    This loader wraps typer-config's toml_loader with JSON schema validation.

    Args:
        config_path: Path to the configuration file

    Returns:
        Validated configuration dictionary

    Raises:
        JsonSchemaValidationError: If config fails schema validation
    """
    global _loaded_config  # pylint: disable=global-statement

    if config_path == '':
        return _loaded_config

    config = toml_loader(config_path)

    schema = _load_schema()
    validator = Draft7Validator(schema)

    errors = sorted(validator.iter_errors(config), key=lambda e: e.path)
    if errors:
        error_messages = []
        for error in errors:
            path = '.'.join(str(p) for p in error.path) if error.path else 'root'
            error_messages.append(f"  {path}: {error.message}")

        raise JsonSchemaValidationError(
            f"Config validation failed for {config_path}:\n" + '\n'.join(error_messages)
        )

    log.debug("Loaded and validated config from file \"%s\"", config_path)

    # Store config globally for access by commands
    _loaded_config = config

    return config


def get_config() -> Config:
    """
    Get the currently loaded configuration - should only be used with prior @use_config(validated_config_callback).

    Returns:
        The loaded configuration dictionary, or empty dict if no config was loaded.
    """
    return _loaded_config


def get_command_config(config: Config, path: str, default: Any = None) -> Any:
    """
    Retrieve command-specific configuration from nested path.

    Args:
        config: Configuration dictionary
        path: Dot-separated path (e.g., 'view.accounts.columns')
        default: Default value if path doesn't exist

    Returns:
        Configuration value at path, or default if not found
    """
    keys = ['commands'] + path.split('.')
    value: Any = config

    for key in keys:
        if isinstance(value, dict):
            value = value.get(key)
            if value is None:
                return default
        else:
            return default

    return value if value is not None else default


# Create the config callback for use with @use_config decorator
validated_config_callback = conf_callback_factory(validated_toml_loader)



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

class InputError(Exception):
    pass

class ValidationError(Exception):
    pass



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import atexit
import locale
from pathlib import Path
from types import SimpleNamespace
import logging
import os
from typing import Optional

from rich import print # pylint: disable=redefined-builtin
from rich.logging import RichHandler
import typer
from typing_extensions import Annotated
from typer_config import use_config

from pp_terminal.output.strategy_factory import create_strategy
from pp_terminal.utils.config import validated_config_callback, get_config
from pp_terminal.exceptions import InputError
from pp_terminal.utils.helper import set_precision
from pp_terminal.output.strategy import OutputFormat
from pp_terminal.utils.plugins import load_command_plugins
from pp_terminal.data.pp_portfolio_builder import PpPortfolioBuilder, CachedPpPortfolioBuilder
from pp_terminal.data.xml_anonymizer import XmlAnonymizer
from . import __version__

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
app.add_typer(typer.Typer(no_args_is_help=True), name="simulate", help="Run simulations on the portfolio data, like share sells or German Vorabpauschale.")
app.add_typer(typer.Typer(no_args_is_help=True), name="view", help="View details about portfolio entities like accounts or securities.")

# init default logging (this is e.g. import for errors during command plugin load)
logging.basicConfig(level=logging.WARN, format="%(message)s", datefmt="[%X]", handlers=[RichHandler(rich_tracebacks=False, show_time=False, show_path=False)])
log = logging.getLogger(__name__)

locale.setlocale(category=locale.LC_ALL, locale='')

# Load external plugins dynamically
load_command_plugins(app)


def version_callback(value: bool) -> None:
    if value:
        print(f"[bold]pp-terminal[/bold] version: {__version__}")
        raise typer.Exit()


def _create_anonymized_temp_file(original_file: Path) -> Path:
    """Create a deterministic anonymized version of the XML file next to the original."""
    temp_path = original_file.parent / f".{original_file.stem}.anon{original_file.suffix}"

    # Use deterministic seed based on file path
    seed = hash(str(original_file.resolve())) % (2**31)
    log.debug("Anonymizing data with seed %d", seed)

    anonymizer = XmlAnonymizer(seed=seed, config=get_config().get('anonymization', {}).get('attributes', {}))
    anonymizer.anonymize_file(original_file, temp_path)
    log.debug("Created anonymized file at %s", temp_path)

    # Register cleanup on program exit
    def cleanup() -> None:
        try:
            os.unlink(temp_path)
            log.debug("Removed temporary anonymized file at %s", temp_path)
        except OSError as e:
            log.warning("Failed to remove temporary file %s: %s", temp_path, e)

    atexit.register(cleanup)

    return temp_path


@app.callback(
    invoke_without_command=True,
    epilog="Small insights today, bigger returns tomorrow.",
    help=f"[bold]pp-terminal[/bold] version {__version__} by [link=https://dev-investor.de]dev-investor[/link]\n\nThe Analytic Companion for Portfolio Performance"
)
@use_config(validated_config_callback)
def main(  # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals
        ctx: typer.Context,
        file: Annotated[Path, typer.Option(help="Portfolio Performance XML file.", show_default=False, exists=True, file_okay=True, dir_okay=False, readable=True)],
        format: OutputFormat = OutputFormat.TABLE,  # pylint: disable=redefined-builtin
        precision: int = 4,
        cache: Annotated[bool, typer.Option('--cache/--no-cache', help='Create cache file for XML.')] = True,
        anonymize: Annotated[bool, typer.Option(help='Anonymize data before processing.')] = False,
        version: Annotated[  # pylint: disable=unused-argument
            Optional[bool],
            typer.Option("--version", callback=version_callback, is_eager=True),  # declared the option name to avoid --no-version
        ] = None,
        verbose: Annotated[Optional[bool], typer.Option('--verbose', help='Enable verbose logging.')] = None,
) -> None:

    if verbose:
        logging.basicConfig(force=True, level=logging.DEBUG, format="%(message)s", datefmt="[%X]", handlers=[RichHandler(rich_tracebacks=True, show_time=False)])

    set_precision(precision)
    source_file = _create_anonymized_temp_file(file) if anonymize else file

    try:
        builder = CachedPpPortfolioBuilder(config=get_config()) if cache else PpPortfolioBuilder(config=get_config())

        ctx.obj = SimpleNamespace(
            source_file=source_file,
            portfolio=builder.construct(source_file),
            output=create_strategy(format),
            config=get_config(),
            verbose=verbose or False)

    except (RuntimeError, InputError) as e:
        if verbose:
            raise e

        log.critical(e)
        raise typer.Abort()


if __name__ == "__main__":
    app()



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass
from typing import Any

import pandas as pd

from pp_terminal.data.filters import filter_not_retired
from pp_terminal.domain.portfolio import Portfolio
from pp_terminal.domain.portfolio_snapshot import PortfolioSnapshot
from pp_terminal.utils.config import get_command_config
from .rules import ValidationRule, create_rule, get_applicable_rules


@dataclass
class ValidationResult:
    """Container for validation results of a single entity."""
    entity_id: str
    violations: list[tuple[ValidationRule, str]]  # (rule, message)

    @property
    def messages(self) -> str:
        """Returns icon-prefixed, semicolon-separated violation messages."""
        if not self.violations:
            return ''

        icon = self._get_icon()
        msgs = '; '.join(msg for _, msg in self.violations)
        return f'{icon} {msgs}' if icon else msgs

    def _get_icon(self) -> str:
        """Returns appropriate icon based on severity."""
        if not self.violations:
            return ''
        has_error = any(rule.is_error() for rule, _ in self.violations)
        return '❌' if has_error else '⚠️'

    @property
    def has_errors(self) -> bool:
        """Returns True if any violation is an error."""
        return any(rule.is_error() for rule, _ in self.violations)

    @classmethod
    def empty(cls, entity_id: str = '') -> 'ValidationResult':
        """Returns empty result with no violations."""
        return cls(entity_id=entity_id, violations=[])


def _validate_entity(
    entity_id: str,
    entity: pd.Series,
    rules: list[ValidationRule],
    context: dict[str, Any]
) -> ValidationResult:
    """Validates single entity against applicable rules."""
    violations = {}

    for rule in get_applicable_rules(entity_id, entity, rules):
        _, message = rule.validate(entity, entity_id, context)
        if message and str(rule) not in violations:  # record only first occurrence for each violation
            violations[str(rule)] = (rule, message)

    return ValidationResult(entity_id=entity_id, violations=list(violations.values()))


def validate_accounts(
    portfolio: Portfolio,
    snapshot: PortfolioSnapshot,
    config: dict[str, Any]
) -> dict[str, ValidationResult]:
    """Validates all deposit accounts. Returns dict mapping account_id -> ValidationResult."""
    rules = [create_rule(rule_config) for rule_config in get_command_config(config, 'validate.accounts.rules', [])]

    total_balances = snapshot.balances.groupby('accountId').sum()
    total_balances.name = 'TotalBalance'

    accounts_with_balances = pd.merge(
        portfolio.deposit_accounts,
        total_balances,
        left_index=True,
        right_index=True,
        how='right',
        validate='one_to_one'
    )

    accounts_with_balances = accounts_with_balances.pipe(filter_not_retired)

    results = {}
    for account_id, account in accounts_with_balances.iterrows():
        context = {
            'balance': account['TotalBalance'],
            'portfolio': portfolio,
            'snapshot': snapshot,
        }
        result = _validate_entity(str(account_id), account, rules, context)
        results[str(account_id)] = result

    return results


def validate_securities(
    portfolio: Portfolio,
    config: dict[str, Any]
) -> dict[str, ValidationResult]:
    """Validates all securities. Returns dict mapping security_id -> ValidationResult."""
    rules = [create_rule(rule_config) for rule_config in get_command_config(config, 'validate.securities.rules', [])]

    latest_prices = portfolio.prices.groupby(['securityId']).tail(1)

    securities_with_prices = pd.merge(
        portfolio.securities,
        latest_prices.reset_index()[['securityId', 'date', 'price']],
        left_index=True,
        right_on='securityId',
        how='left',
        validate='one_to_one'
    ).set_index('securityId')

    securities_with_prices = securities_with_prices.pipe(filter_not_retired)

    results = {}
    for security_id, security in securities_with_prices.iterrows():
        context = {
            'latest_price_date': security.get('date') if pd.notna(security.get('date')) else None,
            'current_price': security.get('price') if pd.notna(security.get('price')) else None,
            'portfolio': portfolio,
            'config': config,
        }
        result = _validate_entity(str(security_id), security, rules, context)
        results[str(security_id)] = result

    return results



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime

import pandas as pd

from pp_terminal.domain.schemas import TransactionType


def filter_earlier_than(df: pd.DataFrame, target_date: datetime) -> pd.DataFrame:
    return df[df.index.get_level_values('date') <= target_date]


def filter_later_than(df: pd.DataFrame, target_date: datetime) -> pd.DataFrame:
    return df[df.index.get_level_values('date') >= target_date]


def filter_by_security(df: pd.DataFrame, security_id: str) -> pd.DataFrame:
    return df[df.index.get_level_values('securityId') == security_id]


def filter_by_account_and_security(df: pd.DataFrame, account_id: str, security_id: str) -> pd.DataFrame:
    return df[(df.index.get_level_values('accountId') == account_id) & (df.index.get_level_values('securityId') == security_id)]


def filter_by_type(df: pd.DataFrame, transaction_types: TransactionType| list[TransactionType]) -> pd.DataFrame:
    if not isinstance(transaction_types, list):
        transaction_types = [transaction_types]

    # we store only the name of the enum to save some space, so we have to convert it here
    cleaned_transaction_types = []
    for transaction_type in transaction_types:
        cleaned_transaction_types.append(transaction_type.name)

    return df[df['type'].isin(cleaned_transaction_types)]


def filter_not_retired(df: pd.DataFrame) -> pd.DataFrame:
    if 'isRetired' not in df.columns:
        return df

    return df[~df['isRetired']]


def drop_empty_values(df: pd.DataFrame | pd.Series) -> pd.DataFrame:
    if df.empty:
        return df

    # Create mask for NaN values and zero values (only for numeric columns)
    if isinstance(df, pd.DataFrame):
        # For DataFrames, handle numeric and non-numeric columns separately
        numeric_cols = df.select_dtypes(include='number').columns
        mask = df.isna()
        if len(numeric_cols) > 0:
            mask = mask | (df[numeric_cols] == 0).reindex(columns=df.columns, fill_value=False)
        df = df[~mask]
    else:
        # For Series, check if numeric before applying == 0
        if pd.api.types.is_numeric_dtype(df):
            df = df[~(df.isna() | (df == 0))]
        else:
            df = df[~df.isna()]

    df.dropna(how='all', axis=0, inplace=True)
    if isinstance(df, pd.DataFrame):
        df.dropna(how='all', axis=1, inplace=True)

    return df


def unstack_column_by_currency(df: pd.DataFrame, column: str, base_currency: str) -> pd.DataFrame:
    column_unstacked = df[column].unstack(level='currency')
    df_modified = df.drop(columns=column).reset_index(level='currency', drop=True)
    df_modified = df_modified[~df_modified.index.get_level_values('accountId').duplicated()]  # drop duplicates ignoring currency

    df_modified = df_modified.join(column_unstacked, how='outer')

    if base_currency in df_modified:
        df_modified.sort_values(by=base_currency, inplace=True)

    return df_modified



import sys
import argparse
import logging
import os.path
import json

from version import __version__
import dbhelper


# uuid to #
sec_map = {}

# uuid to element
output_els = {}

# (portfolio_xact, account_xact) to element
cross_els = {}

xml_id = 0
uuid2xmlid = {}

out = None


class ET_SubElement:

    def __init__(self, parent, tag):
        if parent is None:
            self._indent = 0
        else:
            self._indent = parent._indent + 1
            if not parent.start_written:
                parent.wr_start()
        self.tag = tag
        self.attrib = {}
        self.start_written = False

    def get(self, attr):
        return self.attrib[attr]

    def set(self, attr, val):
        self.attrib[attr] = val

    def indent(self):
        global out
        if self._indent:
            out.write("\n" + "  " * self._indent)

    def _wr_start(self, end):
        global out
        self.indent()
        out.write("<%s" % self.tag)
        for k, v in self.attrib.items():
            out.write(' %s="%s"' % (k, v))
        out.write(end)
        self.start_written = True

    def wr_start(self):
        self._wr_start(">")

    def wr_nb(self):
        self._wr_start("/>")

    def wr_end(self, indent=True):
        global out
        if not self.start_written:
            self.wr_nb()
            return

        if indent:
            self.indent()
            # Still need to indent top closing tag
            if self._indent == 0:
                out.write("\n")
        out.write("</%s>" % self.tag)

    @property
    def text(self):
        raise NotImplementedError

    @text.setter
    def text(self, txt):
        global out
        if txt is None:
            self.wr_nb()
        else:
            self.wr_start()
            out.write(quote_text(txt))
            self.wr_end(False)


def ET_Element(tag):
    return ET_SubElement(None, tag)


# Initial version of this script used lxml.etree (ET) and built a DOM tree
# in memory. But that had poor scalability to large files, so it was rewritten
# using adhoc object mimicking ET API, but writing out most of object content
# to stream eagerly and not keeping unneeded objects in memory.
class ET:
    Element = ET_Element
    SubElement = ET_SubElement


def add_xmlid(el, uuid=None):
    global xml_id

    xml_id += 1
    el.attrib["id"] = str(xml_id)
    if uuid is not None:
        uuid2xmlid[uuid] = str(xml_id)


def ET_SubElementWId(parent, tag, uuid=None):
    el = ET.SubElement(parent, tag)
    add_xmlid(el, uuid)
    return el


def as_bool(v):
    return ["false", "true"][v]


def make_prop(pel, row, prop, row_prop=None, conv=str):
    if row[row_prop or prop] is not None:
        el = ET.SubElement(pel, prop)
        el.text = conv(row[row_prop or prop])


def make_map(pel, rows):
    mapel = ET.SubElement(pel, "map")
    for r in rows:
        enel = ET.SubElement(mapel, "entry")
        ET.SubElement(enel, "string").text = r["attr_uuid"]
        if r["type"] == "limitPrice":
            el = ET.SubElement(enel, r["type"])
            op, val = r["value"].split(" ", 1)
            ET.SubElement(el, "operator").text = op
            ET.SubElement(el, "value").text = val
            el.wr_end()
        elif r["type"] == "bookmark":
            el = ET.SubElement(enel, r["type"])
            if r["value"] is not None:
                subels = json.loads(r["value"])
                for k, v in subels.items():
                    ET.SubElement(el, k).text = v
            el.wr_end()
        else:
            ET.SubElement(enel, r["type"]).text = r["value"]
        enel.wr_end()
    mapel.wr_end()


def make_attributes(pel, rows):
    attributes = ET.SubElement(pel, "attributes")
    make_map(attributes, rows)
    attributes.wr_end()


def make_entry(pel, k, v, type="string"):
    e = ET.SubElement(pel, "entry")
    ET.SubElement(e, "string").text = k
    ET.SubElement(e, type).text = v
    e.wr_end()


def make_configuration(pel, conf):
    conf_el = ET.SubElement(pel, "configuration")
    for k, v in conf.items():
        make_entry(conf_el, k, v)
    conf_el.wr_end()


def make_data_entries(pel, data_rows):
    if data_rows:
        data = ET.SubElement(pel, "data")
        for d_r in data_rows:
            make_entry(data, d_r["name"], d_r["value"], d_r["type"])
        data.wr_end()


def security_ref(uuid, levels=4):
    return uuid2xmlid[uuid]


def make_ref(etree, el, to_el):
    el.set("reference", to_el.get("id"))


def try_ref(etree, el, uuid):
    if uuid in uuid2xmlid:
        el.set("reference", uuid2xmlid[uuid])
        el.wr_nb()
        return True
    return False


def make_xact(etree, pel, tag, xact_r):
            xact = ET.SubElement(pel, tag)
            if try_ref(etree, xact, xact_r["uuid"]):
                return

            add_xmlid(xact, xact_r["uuid"])
            output_els[xact_r["uuid"]] = xact
            make_prop(xact, xact_r, "uuid")
            make_prop(xact, xact_r, "date")
            make_prop(xact, xact_r, "currencyCode", "currency")
            make_prop(xact, xact_r, "amount")
            if xact_r["security"] is not None:
                s = ET.SubElement(xact, "security")
                #s.set("reference", security_ref(xact_r["security"], 5))
                assert try_ref(etree, s, xact_r["security"])

            # 0 or 1
            crit = "from_xact='%s' OR to_xact='%s'" % (xact_r["uuid"], xact_r["uuid"])
            for x_r in dbhelper.select("xact_cross_entry", where=crit):
                #print(dict(x_r))
                x = ET.SubElement(xact, "crossEntry")
                x.set("class", x_r["type"])
                cross_key = (x_r["type"], x_r["from_xact"], x_r["to_xact"])
                existing_x = cross_els.get(cross_key)
                if existing_x is not None:
                    make_ref(etree, x, existing_x)
                    x.wr_nb()
                    continue
                add_xmlid(x)
                cross_els[cross_key] = x
                if x_r["type"] == "account-transfer":
                    accfrom_r = dbhelper.select("account", where="uuid='%s'" % x_r["from_acc"])[0]
                    make_account(etree, x, accfrom_r, el_name="accountFrom")
                    acc_xact_r = dbhelper.select("xact", where="uuid='%s'" % x_r["from_xact"])[0]
                    make_xact(etree, x, "transactionFrom", acc_xact_r)
                    accto_r = dbhelper.select("account", where="uuid='%s'" % x_r["to_acc"])[0]
                    make_account(etree, x, accto_r, el_name="accountTo")
                    acc_xact_r = dbhelper.select("xact", where="uuid='%s'" % x_r["to_xact"])[0]
                    make_xact(etree, x, "transactionTo", acc_xact_r)
                elif x_r["type"] == "portfolio-transfer":
                    accfrom_r = dbhelper.select("account", where="uuid='%s'" % x_r["from_acc"])[0]
                    make_portfolio(etree, x, accfrom_r["uuid"], el_name="portfolioFrom")
                    acc_xact_r = dbhelper.select("xact", where="uuid='%s'" % x_r["from_xact"])[0]
                    make_xact(etree, x, "transactionFrom", acc_xact_r)
                    accto_r = dbhelper.select("account", where="uuid='%s'" % x_r["to_acc"])[0]
                    make_portfolio(etree, x, accto_r["uuid"], el_name="portfolioTo")
                    acc_xact_r = dbhelper.select("xact", where="uuid='%s'" % x_r["to_xact"])[0]
                    make_xact(etree, x, "transactionTo", acc_xact_r)
                else:
                    make_portfolio(etree, x, x_r["from_acc"])
                    port_xact_r = dbhelper.select("xact", where="uuid='%s'" % x_r["from_xact"])[0]
                    make_xact(etree, x, "portfolioTransaction", port_xact_r)

                    acc_r = dbhelper.select("account", where="uuid='%s'" % x_r["to_acc"])[0]
                    make_account(etree, x, acc_r)
                    acc_xact_r = dbhelper.select("xact", where="uuid='%s'" % x_r["to_xact"])[0]
                    make_xact(etree, x, "accountTransaction", acc_xact_r)
                x.wr_end()

            make_prop(xact, xact_r, "shares")
            make_prop(xact, xact_r, "note")
            make_prop(xact, xact_r, "source")

            unit_rows = dbhelper.select("xact_unit", where="xact='%s'" % xact_r["uuid"])
            if unit_rows:
                units = ET.SubElement(xact, "units")
                for unit_r in unit_rows:
                    u = ET.SubElement(units, "unit")
                    u.set("type", unit_r["type"])
                    a = ET.SubElement(u, "amount")
                    a.set("currency", unit_r["currency"])
                    a.set("amount", str(unit_r["amount"]))
                    a.wr_nb()
                    if unit_r["forex_amount"] is not None or unit_r["forex_currency"] is not None:
                        forex = ET.SubElement(u, "forex")
                        forex.set("currency", unit_r["forex_currency"])
                        forex.set("amount", str(unit_r["forex_amount"]))
                        forex.wr_nb()
                        forex = ET.SubElement(u, "exchangeRate")
                        forex.text = str(unit_r["exchangeRate"])
                    u.wr_end()
                units.wr_end()

            make_prop(xact, xact_r, "updatedAt")
            make_prop(xact, xact_r, "type")

            xact.wr_end()


def make_xacts(etree, pel, acc_uuid):
        for xact_r in dbhelper.select("xact", where="account='%s'" % acc_uuid, order="_order"):
            tag = {"account": "account-transaction", "portfolio": "portfolio-transaction"}[xact_r["acctype"]]
            make_xact(etree, pel, tag, xact_r)


def make_portfolio(etree, pel, uuid, el_name="portfolio"):
        el = ET.SubElement(pel, el_name)
        if try_ref(etree, el, uuid):
            return
        add_xmlid(el, uuid)
        output_els[uuid] = el
        port_r = dbhelper.select("account", where="uuid='%s'" % uuid)[0]
        make_prop(el, port_r, "uuid")
        make_prop(el, port_r, "name")
        make_prop(el, port_r, "isRetired", conv=as_bool)
        refacc_r = dbhelper.select("account", where="uuid='%s'" % port_r["referenceAccount"])[0]
        make_account(etree, el, refacc_r, el_name="referenceAccount")

        xacts = ET.SubElement(el, "transactions")
        make_xacts(etree, xacts, port_r["uuid"])
        xacts.wr_end()

        attr_rows = dbhelper.select("account_attr", where="account='%s'" % port_r["uuid"], order="seq")
        make_attributes(el, attr_rows)

        make_prop(el, port_r, "updatedAt")
        el.wr_end()


def make_account(etree, pel, acc_r, el_name="account"):
        acc = ET.SubElement(pel, el_name)
        if try_ref(etree, acc, acc_r["uuid"]):
            return
        add_xmlid(acc, acc_r["uuid"])
        output_els[acc_r["uuid"]] = acc
        make_prop(acc, acc_r, "uuid")
        make_prop(acc, acc_r, "name")
        make_prop(acc, acc_r, "currencyCode", "currency")
        make_prop(acc, acc_r, "note")
        make_prop(acc, acc_r, "isRetired", conv=as_bool)

        xacts = ET.SubElement(acc, "transactions")
        make_xacts(etree, xacts, acc_r["uuid"])
        xacts.wr_end()

        attr_rows = dbhelper.select("account_attr", where="account='%s'" % acc_r["uuid"], order="seq")
        make_attributes(acc, attr_rows)

        make_prop(acc, acc_r, "updatedAt")
        acc.wr_end()


def quote_text(s):
    pats = (
        ("&", "&amp;"),
        ("<", "&lt;"),
        (">", "&gt;"),
        ('"', "&quot;"),
        ("'", "&apos;"),
    )
    for p, r in pats:
        s = s.replace(p, r)
    return s


def make_taxonomy_level(etree, pel, level_r):
        tag = "root" if level_r["parent"] is None else "classification"
        level = ET_SubElementWId(pel, tag, level_r["uuid"])
        output_els[level_r["uuid"]] = level
        make_prop(level, level_r, "id", "uuid")
        make_prop(level, level_r, "name")
        make_prop(level, level_r, "color")

        if level_r["parent"]:
            p = ET.SubElement(level, "parent")
            assert try_ref(etree, p, level_r["parent"])

        chds = ET.SubElement(level, "children")
        for e_r in dbhelper.select("taxonomy_category", where="parent='%s'" % level_r["uuid"]):
            make_taxonomy_level(etree, chds, e_r)
        chds.wr_end()

        assgn = ET.SubElement(level, "assignments")
        for a_r in dbhelper.select("taxonomy_assignment", where="category='%s'" % level_r["uuid"]):
            a = ET.SubElement(assgn, "assignment")
            iv = ET.SubElement(a, "investmentVehicle")
            iv.set("class", a_r["item_type"])
            assert try_ref(etree, iv, a_r["item"])
            make_prop(a, a_r, "weight")
            make_prop(a, a_r, "rank")

            data_rows =  dbhelper.select("taxonomy_assignment_data", where="assignment=%d" % a_r["_id"])
            make_data_entries(a, data_rows)
            a.wr_end()
        assgn.wr_end()

        make_prop(level, level_r, "weight")
        make_prop(level, level_r, "rank")

        data_rows = dbhelper.select("taxonomy_data", where="category='%s'" % level_r["uuid"])
        make_data_entries(level, data_rows)
        level.wr_end()


def main():
    global out
    out = sys.stdout
    if args.xml_file:
        out = open(args.xml_file, "w", encoding="utf-8", newline="\n")

    root = ET.Element("client")
    add_xmlid(root)
    etree = None
    for n in ["version", "baseCurrency"]:
        row = dbhelper.select("property", where="name='%s'" % n)[0]
        ET.SubElement(root, n).text = row["value"]

    securities = ET.SubElement(root, "securities")

    for i, sec_r in enumerate(dbhelper.select("security")):
    #    print(dict(sec_r))
        sec_map[sec_r["uuid"]] = i
        sec = ET_SubElementWId(securities, "security", sec_r["uuid"])
        output_els[sec_r["uuid"]] = sec
        make_prop(sec, sec_r, "uuid")
        make_prop(sec, sec_r, "onlineId")
        make_prop(sec, sec_r, "name")
        make_prop(sec, sec_r, "currencyCode", "currency")
        make_prop(sec, sec_r, "targetCurrencyCode", "targetCurrency")
        make_prop(sec, sec_r, "note")
        make_prop(sec, sec_r, "isin")
        make_prop(sec, sec_r, "tickerSymbol")
        make_prop(sec, sec_r, "calendar")
        make_prop(sec, sec_r, "wkn")
        make_prop(sec, sec_r, "feedTickerSymbol")
        make_prop(sec, sec_r, "feed")
        make_prop(sec, sec_r, "feedURL")

        prices = ET.SubElement(sec, "prices")
        for price_r in dbhelper.select("price", where="security='%s'" % sec_r["uuid"]):
            p = ET.SubElement(prices, "price")
            p.set("t", price_r["tstamp"])
            p.set("v", str(price_r["value"]))
            p.wr_nb()
        prices.wr_end()

        make_prop(sec, sec_r, "latestFeed")
        make_prop(sec, sec_r, "latestFeedURL")

        # Can be 0 or 1
        for latest_r in dbhelper.select("latest_price", where="security='%s'" % sec_r["uuid"]):
            latest = ET.SubElement(sec, "latest")
            latest.set("t", latest_r["tstamp"])
            latest.set("v", str(latest_r["value"]))
            make_prop(latest, latest_r, "high")
            make_prop(latest, latest_r, "low")
            make_prop(latest, latest_r, "volume")
            latest.wr_end()

        attr_rows = dbhelper.select("security_attr", where="security='%s'" % sec_r["uuid"], order="seq")
        make_attributes(sec, attr_rows)

        events = ET.SubElement(sec, "events")
        order_args = {}
        if args.sort_events:
            order_args = {"order": "date, details"}
        for event_r in dbhelper.select("security_event", where="security='%s'" % sec_r["uuid"], **order_args):
            event = ET.SubElement(events, "event")
            make_prop(event, event_r, "date")
            make_prop(event, event_r, "type")
            make_prop(event, event_r, "details")
            event.wr_end()
        events.wr_end()

        for prop_r in dbhelper.select("security_prop", where="security='%s'" % sec_r["uuid"], order="seq"):
            p = ET.SubElement(sec, "property")
            p.set("type", prop_r["type"])
            p.set("name", prop_r["name"])
            p.text = prop_r["value"]

        make_prop(sec, sec_r, "isRetired", conv=as_bool)
        make_prop(sec, sec_r, "updatedAt")
        sec.wr_end()
    securities.wr_end()

    watchlists = ET.SubElement(root, "watchlists")
    for wlist_r in dbhelper.select("watchlist"):
        wlist = ET_SubElementWId(watchlists, "watchlist")
        make_prop(wlist, wlist_r, "name")
        secs = ET.SubElement(wlist, "securities")
        for wlist_sec_r in dbhelper.select("watchlist_security", where="list=%d" % wlist_r["_id"]):
            s = ET.SubElement(secs, "security")
            s.set("reference", security_ref(wlist_sec_r["security"]))
            s.wr_nb()
        secs.wr_end()
        wlist.wr_end()
    watchlists.wr_end()

    accounts = ET.SubElement(root, "accounts")
    for acc_r in dbhelper.select("account", where="type='account'", order="_order"):
        make_account(etree, accounts, acc_r)
    accounts.wr_end()

    portfolios = ET.SubElement(root, "portfolios")
    for acc_r in dbhelper.select("account", where="type='portfolio'", order="_order"):
        make_portfolio(etree, portfolios, acc_r["uuid"])
    portfolios.wr_end()

    plans = ET.SubElement(root, "plans")
    plans.wr_end()

    taxonomies = ET.SubElement(root, "taxonomies")
    for taxon_r in dbhelper.select("taxonomy"):
        taxon = ET.SubElement(taxonomies, "taxonomy")
        make_prop(taxon, taxon_r, "id", "uuid")
        make_prop(taxon, taxon_r, "name")
        taxon_dim_rows =  dbhelper.select("taxonomy_data", where="taxonomy='%s' AND category IS NULL AND name='dimension'" % taxon_r["uuid"])
        if taxon_dim_rows:
            el = ET.SubElement(taxon, "dimensions")
            for taxon_dim_r in taxon_dim_rows:
                ET.SubElement(el, "string").text = taxon_dim_r["value"]
            el.wr_end()
        e_r = dbhelper.select("taxonomy_category", where="uuid='%s'" % taxon_r["root"])[0]
        make_taxonomy_level(etree, taxon, e_r)
        taxon.wr_end()
    taxonomies.wr_end()


    dashboards = ET.SubElement(root, "dashboards")
    for dashb_r in dbhelper.select("dashboard"):
        dashb = ET.SubElement(dashboards, "dashboard")
        dashb.set("name", dashb_r["name"])
        make_prop(dashb, dashb_r, "id")
        make_configuration(dashb, json.loads(dashb_r["config_json"]))
        columns = ET.SubElement(dashb, "columns")
        for col_j in json.loads(dashb_r["columns_json"]):
            col = ET.SubElement(columns, "column")
            make_prop(col, col_j, "weight")
            widgets = ET.SubElement(col, "widgets")
            for wid_j in col_j["widgets"]:
                wid = ET.SubElement(widgets, "widget")
                wid.set("type", wid_j["type"])
                make_prop(wid, wid_j, "label")
                if "config" in wid_j:
                    make_configuration(wid, wid_j["config"])
                wid.wr_end()
            widgets.wr_end()
            col.wr_end()
        columns.wr_end()
        dashb.wr_end()
    dashboards.wr_end()


    properties = ET.SubElement(root, "properties")
    for prop_r in dbhelper.select("property", where="special=0"):
        make_entry(properties, prop_r["name"], prop_r["value"])
    properties.wr_end()

    settings = ET.SubElement(root, "settings")

    bookmarks = ET.SubElement(settings, "bookmarks")
    for bmark_r in dbhelper.select("bookmark"):
        bmark = ET.SubElement(bookmarks, "bookmark")
        make_prop(bmark, bmark_r, "label")
        make_prop(bmark, bmark_r, "pattern")
        bmark.wr_end()
    bookmarks.wr_end()

    attrtypes = ET.SubElement(settings, "attributeTypes")
    for attr_type_r in dbhelper.select("attribute_type"):
        attr_type = ET.SubElement(attrtypes, "attribute-type")
        make_prop(attr_type, attr_type_r, "id")
        make_prop(attr_type, attr_type_r, "name")
        make_prop(attr_type, attr_type_r, "columnLabel")
        make_prop(attr_type, attr_type_r, "source")
        make_prop(attr_type, attr_type_r, "target")
        make_prop(attr_type, attr_type_r, "type")
        make_prop(attr_type, attr_type_r, "converterClass")
        prop_list = json.loads(attr_type_r["props_json"])
        if prop_list:
            props = ET.SubElement(attr_type, "properties")
            for p in prop_list:
                make_entry(props, p["name"], p["value"], type=p["type"])
            props.wr_end()
        elif attr_type_r["type"].rsplit(".", 1)[1] in ("LimitPrice",):
            props = ET.SubElement(attr_type, "properties")
            props.wr_end()
        attr_type.wr_end()
    attrtypes.wr_end()

    config_sets = ET.SubElement(settings, "configurationSets")
    for cset_r in dbhelper.select("config_set"):
        el = ET.SubElement(config_sets, "entry")
        make_prop(el, cset_r, "string", "name")
        el2 = ET.SubElement(el, "config-set")
        el3 = ET.SubElement(el2, "configurations")
        for centry_r in dbhelper.select("config_entry", where="config_set=%d" % cset_r["_id"]):
            centry = ET.SubElement(el3, "config")
            make_prop(centry, centry_r, "uuid")
            make_prop(centry, centry_r, "name")
            make_prop(centry, centry_r, "data")
            centry.wr_end()
        el3.wr_end()
        el2.wr_end()
        el.wr_end()
    config_sets.wr_end()

    settings.wr_end()

    root.wr_end()


if __name__ == "__main__":
    argp = argparse.ArgumentParser(description="Export Sqlite DB to PortfolioPerformance XML file")
    argp.add_argument("db", help="input DB (filename/connect string)")
    argp.add_argument("xml_file", nargs="?", help="output XML file (stdout if not provided)")
    argp.add_argument("--dbtype", choices=("sqlite", "pgsql"), default="sqlite", help="select database type")
    argp.add_argument("--sort-events", action="store_true", help="sort events by date (then description)")
    argp.add_argument("--debug", action="store_true", help="enable debug logging")
    argp.add_argument("--version", action="version", version="%(prog)s " + __version__)
    args = argp.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    dbhelper.init(args.dbtype, args.db)
    main()



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from datetime import datetime
from typing import Annotated, Any, cast

import numpy as np
import pandas as pd
import typer
from pandera.typing import DataFrame
from rich.console import Console

from pp_terminal.data.filters import filter_later_than, filter_by_type
from pp_terminal.utils.helper import get_last_year, footer
from pp_terminal.output.strategy import OutputStrategy
from pp_terminal.domain.portfolio import Portfolio
from pp_terminal.domain.portfolio_snapshot import PortfolioSnapshot
from pp_terminal.domain.schemas import Percent, TransactionType, Money, InterestResultSchema
from pp_terminal.output.table_decorator import TableOptions, format_value

app = typer.Typer()
console = Console()
log = logging.getLogger(__name__)


_DAYS_PER_YEAR = 360


def calculate_interest(snapshot_begin: PortfolioSnapshot, snapshot_end: PortfolioSnapshot, interest_rate: Percent) -> DataFrame[InterestResultSchema]:
    transactions = snapshot_end.deposit_account_transactions

    df = transactions.reset_index(level='date').sort_values(by=['accountId', 'currency', 'date'])
    df = df.groupby(['accountId', 'currency', 'date']).agg({'date': 'min', 'amount': 'sum'})

    df['days_diff'] = df.groupby(['accountId', 'currency'])['date'].diff().dt.days.fillna(0).astype(int)
    df['balance'] = df.groupby(['accountId', 'currency'])['amount'].cumsum()
    df['weighted_balance'] = df['balance'] * df['days_diff']
    df['interest_rate'] = np.power(1 + interest_rate / 100 / _DAYS_PER_YEAR, df['days_diff']) - 1
    df['interest'] = df['balance'] * df['interest_rate']

    df = df.pipe(filter_later_than, target_date=snapshot_begin.date)
    df = df.groupby(['accountId', 'currency']).agg({'interest': 'sum', 'weighted_balance': 'sum', 'days_diff': 'sum'})
    df['mean_balance'] = (df['weighted_balance'] / df['days_diff']).fillna(0)

    df = df.reset_index(level='currency')

    interest_transactions = (transactions.pipe(filter_by_type, transaction_types=[TransactionType.INTEREST, TransactionType.INTEREST_CHARGE])
                             .pipe(filter_later_than, target_date=snapshot_begin.date)
                             .reset_index(level='date'))
    interest_transactions['amount'] = interest_transactions['amount'] + interest_transactions['taxes']

    actual_interest_df = interest_transactions.groupby(['accountId', 'currency'])['amount'].sum().fillna(0).reset_index(level='currency')
    actual_interest_df = actual_interest_df.rename(columns={'amount': 'actual_interest'})

    # Merge actual interest into main dataframe
    df = df.merge(actual_interest_df, left_index=True, right_index=True, how='left', suffixes=('', '_drop'))
    # Handle potential duplicate currency columns from merge
    if 'currency_drop' in df.columns:
        df = df.drop(columns=['currency_drop'])

    # Merge with account names - deposit_accounts also has currency, so we'll get currency_x and currency_y
    interest_per_account = (pd.merge(snapshot_end.portfolio.deposit_accounts, df, left_index=True, right_index=True, how="right", validate='one_to_one', suffixes=('_account', ''))
                .sort_values(by='interest'))

    # Keep the currency from df (interest calculations), drop the one from accounts
    if 'currency_account' in interest_per_account.columns:
        interest_per_account = interest_per_account.drop(columns=['currency_account'])

    result = interest_per_account[interest_per_account['interest'] > 0][['name', 'currency', 'mean_balance', 'interest', 'actual_interest']]
    return InterestResultSchema.validate(result)


def _format_value_wrapper(value: Any, index: str, row: pd.Series) -> str:
    if index == 'Actual Interest' and isinstance(value, Money):
        color = 'red' if value < row['Simulated Interest'] else 'green'
        return f"[{color}]{format_value(value, index, row)}[/{color}]"

    return format_value(value, index, row)


@app.command(name="interest")
def simulate_interest_rate(
        ctx: typer.Context,
        interest_rate: Annotated[Percent, typer.Option(min=0, max=100, prompt="Interest Rate (%)", prompt_required=True)],
        year: Annotated[datetime, typer.Option(formats=["%Y"], help="The year to calculate the preliminary tax for", prompt=True, default_factory=get_last_year)],
) -> None:
    """
    Simulate a given interest rate for a deposit account.
    """

    portfolio = cast(Portfolio, ctx.obj.portfolio)
    output = cast(OutputStrategy, ctx.obj.output)

    snapshot_begin = PortfolioSnapshot(portfolio, datetime(year.year, 1, 1))
    snapshot_end = PortfolioSnapshot(portfolio, datetime(year.year, 12, 31))

    df = calculate_interest(snapshot_begin, snapshot_end, interest_rate)
    if df is not None:
        df = df.rename(columns={'mean_balance': '⌀ Balance', 'interest': 'Simulated Interest', 'actual_interest': 'Actual Interest'})
        df.insert(3, 'Interest Rate', f"{interest_rate}%")

    console.print(*output.result_table(
        df, TableOptions(title='Simulated Interest on Accounts', caption=f"for {year.strftime("%Y")}, excl. taxes", show_index=False, show_total=False, value_formatter=_format_value_wrapper)
    ))
    console.print(output.text(footer()), style="dim")



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from datetime import datetime
from pathlib import Path

import pandas as pd
import typer
from pandera.errors import SchemaError
from pandera.typing import DataFrame

from pp_terminal.exceptions import InputError
from pp_terminal.output.strategy import Console
from pp_terminal.domain.schemas import TaxPaidSchema, Percent, TaxLotSchema

app = typer.Typer()
console = Console()
log = logging.getLogger(__name__)


def load_prepaid_tax_data_from_csv(csv_path: Path, tax_rate: Percent) -> DataFrame[TaxPaidSchema]:
    """
    Load paid tax data from CSV (e.g. "Vorabpauschale").
    """
    log.debug('Loading paid tax data from "%s"', csv_path)

    try:
        df = pd.read_csv(csv_path, sep=';', parse_dates=['date'])
    except FileNotFoundError as e:
        raise InputError(f"Prepaid tax data CSV file not found: {csv_path}") from e
    except Exception as e:
        raise InputError(f"Failed to read prepaid tax data CSV: {e}") from e

    df['year'] = df['date'].dt.year
    df['tax_per_share'] = df['deemed_income_base_per_share'] * tax_rate
    df = df.set_index(['year', 'account_id', 'security_id'])

    try:
        return TaxPaidSchema.validate(df[['tax_per_share', 'tax_free_allowance']])
    except SchemaError as e:
        raise InputError(f"Prepaid tax data CSV is missing required columns: {e}") from e


def calculate_prepaid_tax_per_lot(
    lots: DataFrame[TaxLotSchema],
    current_date: datetime,
    tax_csv_data: DataFrame[TaxPaidSchema] | None
) -> pd.Series:
    """
    Calculate prepaid tax per lot (e.g. Vorabpauschale).

    Args:
        lots: Current FIFO lots DataFrame (after matching sales)
        current_date: Evaluation date (for year range calculation)
        tax_csv_data: CSV data with taxes paid per share, indexed by (year, account_id, security_id)

    Returns:
        Series of prepaid tax amounts, indexed to match the input lots dataframe.
    """
    if lots.empty or tax_csv_data is None or tax_csv_data.empty:
        return pd.Series(0.0, index=lots.index)

    # Add year columns to track original lot
    lots_with_years = lots.copy().reset_index()
    lots_with_years['first_year'] = lots_with_years['date'].dt.year
    lots_with_years['last_year'] = current_date.year - 1

    # Filter out current year purchases (no prepaid tax yet)
    lots_with_years = lots_with_years[lots_with_years['last_year'] >= lots_with_years['first_year']]

    if lots_with_years.empty:
        return pd.Series(0.0, index=lots.index)

    # Create year range for each lot
    lots_with_years['year_range'] = lots_with_years.apply(
        lambda row: list(range(row['first_year'], row['last_year'] + 1)),
        axis=1
    )

    # Explode into lot-year pairs
    lot_years = lots_with_years.explode('year_range').rename(columns={'year_range': 'year'})

    # Calculate month_factor (prorate for purchase year)
    lot_years['month_factor'] = 1.0
    is_first_year = lot_years['year'] == lot_years['first_year']
    lot_years.loc[is_first_year, 'month_factor'] = (
        (13 - lot_years.loc[is_first_year, 'date'].dt.month) / 12.0
    )

    tax_csv_data = tax_csv_data.reset_index().rename(columns={'account_id': 'accountId', 'security_id': 'securityId'})  # @todo
    lot_years = lot_years.merge(
        tax_csv_data,
        on=['year', 'accountId', 'securityId'],
        how='inner'
    )

    if lot_years.empty:
        return pd.Series(0.0, index=lots.index)

    if 'tax_free_allowance' not in lot_years.columns:
        lot_years['tax_free_allowance'] = 0.0
    else:
        lot_years['tax_free_allowance'] = lot_years['tax_free_allowance'].fillna(0.0)

    lot_years['taxes'] = (
        lot_years['shares'] * lot_years['tax_per_share'] * lot_years['month_factor'] - lot_years['tax_free_allowance']
    )

    # Group by original lot identity (MultiIndex columns) to sum taxes across all years
    per_lot_tax = lot_years.groupby(['date', 'accountId', 'securityId'])['taxes'].sum()

    # Reindex to match original lots dataframe, filling missing with 0
    return per_lot_tax.reindex(lots.index, fill_value=0.0)
